"""In WFS, a "projection" is placed on top of the queried data.

It translates the incoming data into the subset of properties to display.
Practically, this code does inform and adjust the constructed QuerySet
to make sure it will provide only the actual fields that are part of the projection.
"""

from __future__ import annotations

import itertools
import logging
import operator
import typing
from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass
from functools import cached_property

from gisserver.types import (
    GeometryXsdElement,
    XPathMatch,
    XsdElement,
    XsdNode,
    XsdTypes,
    _XsdElement_WithComplexType,
)

if typing.TYPE_CHECKING:
    from django.contrib.gis.geos import GEOSGeometry
    from django.db import models

    from gisserver.crs import CRS
    from gisserver.features import FeatureType
    from gisserver.parsers import fes20, wfs20

__all__ = (
    "FeatureProjection",
    "FeatureRelation",
)

logger = logging.getLogger(__name__)


class FeatureProjection:
    """Tell which fields to access and render for a single feature.
    This is inspired on 'fes:AbstractProjectionClause'.

    Instead of walking over the full XSD object tree,
    this object wraps that and makes sure only the actual requested fields are used.
    When a ``PROPERTYNAME`` (or ``<wfs:PropertyName>``) is used in the request,
    this will limit which fields to retrieve, which to prefetch, and which to render.
    """

    #: Referencing the Feature that is rendered.
    feature_type: FeatureType

    #: The list of root element to render for this feature.
    xsd_root_elements: list[XsdElement]

    #: The subset of child nodes to render for a given element.
    xsd_child_nodes: dict[XsdElement | None, list[XsdElement]]

    #: The output Coordinate Reference System
    output_crs: CRS

    #: Whether the output should be rendered without wrapper tags (for GetFeatureById).
    output_standalone: bool

    def __init__(
        self,
        feature_types: list[FeatureType],
        property_names: list[wfs20.PropertyName] | None = None,
        value_reference: fes20.ValueReference | None = None,
        output_crs: CRS | None = None,
        output_standalone: bool = False,
    ):
        """

        :param feature_types: The feature types used by this query. Typically one, unless there is a JOIN.
        :param property_names: Limited list of fields to render only.
        :param value_reference: Single element to display fo GetPropertyValue
        :param output_crs: Which coordinate reference system to use for geometry data.
        :param output_standalone: Used for the ``GetFeatureById`` stored query.
            This removes the ``wfs:FeatureCollection><wfs:member>`` wrapper elements from the output.
        """
        self.feature_types = feature_types
        self.feature_type = feature_types[0]  # JOIN still not supported.
        self.property_names = property_names
        self.value_reference = value_reference
        self.output_crs: CRS = output_crs or self.feature_type.crs
        self.output_standalone = output_standalone  # for GetFeatureById
        self._extra_matches = []

        if property_names:
            # Only a selection of the tree will be rendered.
            # Discover which elements should be rendered.
            child_nodes = self._get_child_nodes_subset()
            self.xsd_root_elements = child_nodes.pop(None)  # Pop it to avoid recursion risks
            self.xsd_child_nodes = child_nodes
        else:
            # All elements of the tree will be rendered, retrieve all elements.
            # Note xsd_child_nodes can be altered below, so don't assign it a cached property.
            self.xsd_root_elements = self.feature_type.xsd_type.elements_including_base
            self.xsd_child_nodes = {
                node: node.type.elements
                for node in self.feature_type.xsd_type.all_complex_elements
            }

    def _get_child_nodes_subset(self) -> dict[XsdElement | None, list[XsdElement]]:
        """Translate the PROPERTYNAME into a dictionary of nodes to render."""
        child_nodes = defaultdict(list)
        for xpath_match in self.xpath_matches:
            # Make sure the tree is known beforehand, and it has no duplicate parent nodes.
            # (e.g. PROPERTYNAME=node/child1,node/child2)
            parent = None
            for xsd_node in xpath_match.nodes:
                if xsd_node not in child_nodes[parent]:
                    child_nodes[parent].append(xsd_node)
                parent = xsd_node

        return dict(child_nodes)

    def add_field(self, xsd_element: XsdElement):
        """Restore the retrieval of a field that was not asked for in the original query."""
        if xsd_element.type.is_complex_type:
            raise NotImplementedError("Can't restore nested elements right now")
        if not self.property_names or xsd_element in self.xsd_root_elements:
            return

        logger.debug("Projection: added %s", xsd_element)
        self.xsd_root_elements.append(xsd_element)
        self._extra_matches.append(XPathMatch(self.feature_type, [xsd_element], ""))
        _clear_cached_properties(self)

    def remove_fields(self, predicate: Callable[[XsdElement], bool]):
        """Remove elements from the projection based on a given rule.
        This helps to remove M2M and Array elements for CSV output for example.

        Make sure this function is called as early as possible,
        before other logic already read these attributes.
        """
        self.xsd_root_elements, removed_nodes = _partition(predicate, self.xsd_root_elements)

        for xsd_child_root, xsd_child_elements in self.xsd_child_nodes.items():
            if xsd_child_root in removed_nodes:
                continue  # already removed the top-level, don't bother checking here

            keep_children, remove_children = _partition(predicate, xsd_child_elements)
            self.xsd_child_nodes[xsd_child_root] = keep_children
            removed_nodes.update(remove_children)

        # Remove complete trees if their parent was removed.
        for xsd_child_root in list(self.xsd_child_nodes):
            if xsd_child_root in removed_nodes:
                self.xsd_child_nodes.pop(xsd_child_root, None)

        if logger.isEnabledFor(logging.DEBUG) and removed_nodes:
            logger.debug("Projection: removed %s", removed_nodes)

        _clear_cached_properties(self)

    @cached_property
    def _geometry_getter(self):
        return operator.attrgetter(self.main_geometry_element.orm_path)

    def get_main_geometry_value(self, instance: models.Model) -> GEOSGeometry | None:
        """Efficiently retrieve the value for the main geometry element."""
        if self.main_geometry_element is None:
            return None
        else:
            return self._geometry_getter(instance)

    @cached_property
    def xpath_matches(self) -> list[XPathMatch]:
        """Resolve which elements the property names point to"""
        if not self.property_names:
            raise RuntimeError("This method is only useful for propertyname projections.")

        return [
            self.feature_type.resolve_element(property_name.xpath, property_name.xpath_ns_aliases)
            for property_name in self.property_names
        ] + self._extra_matches

    @cached_property
    def all_elements(self) -> list[XsdElement]:
        """Return ALL elements of all levels to render."""
        return self.xsd_root_elements + list(
            itertools.chain.from_iterable(self.xsd_child_nodes.values())
        )

    @cached_property
    def all_geometry_elements(self) -> list[GeometryXsdElement]:
        """Tell which GML elements will be hit."""
        return [e for e in self.all_elements if e.type.is_geometry]

    @cached_property
    def all_complex_elements(self) -> list[_XsdElement_WithComplexType]:
        """Return ALL tree elements with a complex type, including child elements with a complex types."""
        return list(self.xsd_child_nodes.keys())

    @cached_property
    def all_flattened_elements(self) -> list[XsdElement]:
        """Shortcut to get ALL tree elements with a flattened model attribute"""
        if not self.property_names:
            return self.feature_type.xsd_type.flattened_elements
        else:
            return [e for e in self.all_elements if e.is_flattened]

    @cached_property
    def has_bounded_by(self) -> bool:
        """Tell whether the <gml:boundedBy> element is included for rendering."""
        return any(e.type is XsdTypes.gmlBoundingShapeType for e in self.xsd_root_elements)

    @cached_property
    def main_geometry_element(self) -> GeometryXsdElement | None:
        """Return the field used to describe the geometry of the feature.
        When the projection excludes the geometry, ``None`` is returned.
        """
        geo_element = self.feature_type.main_geometry_element
        if self.property_names and geo_element not in self.all_elements:
            return None

        return geo_element

    @cached_property
    def geometry_elements(self) -> list[GeometryXsdElement]:
        """Tell which GML elements will be hit at the root-level."""
        return [
            e
            for e in self.xsd_root_elements
            if e.type.is_geometry and e.type is not XsdTypes.gmlBoundingShapeType
        ]

    @cached_property
    def property_value_node(self) -> XsdNode:
        """For GetPropertyValue, resolve the element that is rendered."""
        if self.value_reference is None:
            raise RuntimeError("This method is only useful for GetPropertyValue calls.")
        return (
            self.feature_types[0]
            .resolve_element(self.value_reference.xpath, self.value_reference.xpath_ns_aliases)
            .child
        )

    @cached_property
    def orm_relations(self) -> list[FeatureRelation]:
        """Tell which fields will be retrieved from related fields.

        This gives an object layout based on the XSD elements,
        that can be used for prefetching data.
        """
        related_models: dict[str, type[models.Model]] = {}
        fields: dict[str, set[XsdElement]] = defaultdict(set)
        elements = defaultdict(list)

        # Check all elements that render as "dotted" flattened relation
        for xsd_element in self.all_flattened_elements:
            if xsd_element.source is not None:
                # Split "relation.field" notation into path, and take the field as child attribute.
                obj_path, field = xsd_element.orm_relation
                elements[obj_path].append(xsd_element)
                fields[obj_path].add(xsd_element)
                # field is already on relation:
                related_models[obj_path] = xsd_element.source.model

        # Check all elements that render as "nested" complex type:
        for xsd_element in self.all_complex_elements:
            # The complex element itself points to the root of the path,
            # all sub elements become the child attributes.
            obj_path = xsd_element.orm_path
            elements[obj_path].append(xsd_element)
            fields[obj_path] = {
                f
                for f in self.xsd_child_nodes[xsd_element]
                if not f.is_many or f.is_array  # exclude M2M, but include ArrayField
            }
            if xsd_element.source:
                # field references a related object:
                related_models[obj_path] = xsd_element.source.related_model

        return [
            FeatureRelation(
                orm_path=obj_path,
                sub_fields=sub_fields,
                related_model=related_models.get(obj_path),
                xsd_elements=elements[obj_path],
            )
            for obj_path, sub_fields in fields.items()
        ]

    @cached_property
    def only_fields(self) -> list[str]:
        """Tell which fields to limit the queryset to.
        This excludes M2M fields because those are not part of the local model data.
        """
        if self.property_names is not None:
            return [
                # TODO: While ORM rel__child paths can be passed to .only(),
                # these may not be accurately applied to foreign keys yet.
                # Also, this is bypassed by our generated Prefetch() objects.
                xpath_match.orm_path
                for xpath_match in self.xpath_matches
                if not xpath_match.is_many or xpath_match.child.is_array
            ]
        else:
            # Also limit the queryset to the actual fields that are shown.
            # No need to request more data
            return [
                f.orm_field
                for f in self.feature_type.xsd_type.elements
                if not f.is_many or f.is_array  # avoid M2M fields for .only(), but keep ArrayField
            ]


@dataclass
class FeatureRelation:
    """Tell which related fields are queried by the feature."""

    #: The ORM path that is queried for this particular relation
    orm_path: str
    #: The fields that will be retrieved for that path (limited by the projection)
    sub_fields: set[XsdElement]
    #: The model that is accessed for this relation (if set)
    related_model: type[models.Model] | None
    #: The source elements that access this relation. Could be multiple for flattened relations.
    xsd_elements: list[XsdElement]

    @cached_property
    def _local_model_field_names(self) -> list[str]:
        """Tell which local fields of the model will be accessed by this feature."""
        return [
            model_field.name
            for field in self.sub_fields
            if not (model_field := field.source).many_to_many and not model_field.one_to_many
        ] + self._local_backlink_field_names

    @property
    def _local_backlink_field_names(self) -> list[str]:
        # When this relation is retrieved through a ManyToOneRel (reverse FK),
        # the prefetch_related() also needs to have the original foreign key
        # in order to link all prefetches to the proper parent instance.
        return [
            xsd_element.source.field.name
            for xsd_element in self.xsd_elements
            if xsd_element.source is not None and xsd_element.source.one_to_many
        ]

    @property
    def geometry_elements(self) -> list[GeometryXsdElement]:
        """Tell which geometry elements this relation will access."""
        return [f for f in self.sub_fields if f.type.is_geometry]


def _partition(predicate, items: list) -> tuple[list, set]:
    """Semi-efficient way to split a list into items that match/don't match the condition."""
    # more_itertools.partition() is faster, but that can be neglected with a short list.
    return list(itertools.filterfalse(predicate, items)), set(filter(predicate, items))


def _clear_cached_properties(object):
    """Remove the caches from the cached_property decorator on an object."""
    cls = object.__class__
    for property_name in list(object.__dict__):
        if (prop := getattr(cls, property_name, None)) is not None and isinstance(
            prop, cached_property
        ):
            del object.__dict__[property_name]



"""Projection tag parsing for WFS 2.0"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from gisserver.exceptions import (
    InvalidParameterValue,
    OperationParsingFailed,
    OperationProcessingFailed,
)
from gisserver.parsers.ast import AstNode, tag_registry
from gisserver.parsers.query import CompiledQuery
from gisserver.parsers.xml import NSElement, xmlns
from gisserver.types import XPathMatch


class ResolveValue(Enum):
    """The ``wfs:ResolveValueType`` enum, used by :class:`StandardResolveParameters`."""

    local = "local"
    remote = "remote"
    all = "all"
    none = "none"

    @classmethod
    def _missing_(cls, value):
        raise OperationParsingFailed(f"Invalid resolve value: {value}", locator="resolve")


@dataclass
@tag_registry.register("PropertyName", xmlns.wfs)
class PropertyName(AstNode):
    """The ``<wfs:PropertyName>`` element in the projection clause.

    This parses and handles the syntax::

        <wfs:PropertyName>ns0:tagname</wfs:PropertyName>

    More advanced syntax is not supported, such as::

      <wfs:PropertyName resolve="all" resolvePath="valueOf(relatedTo)">valueOf(registerEntry)</wfs:PropertyName>

    Note this element exists in the WFS namespace, not the FES namespace!
    The ``<fes:PropertyName>`` is an old FES 1.x element that has been replaced by ``<fes:ValueReference>``.
    The old ``<fes:PropertyName>`` is still supported as an alias by this server.
    """

    xpath: str  # Officially only a 'QName'
    xpath_ns_aliases: dict[str, str]

    # Unused, yet included for completeness.
    # These are available in the XML syntax, and override
    # the default given in the <wfs:GetFeature> element.
    resolve: ResolveValue = ResolveValue.none
    resolve_depth: int | None = None
    resolve_path: str | None = None
    resolve_timeout: int = 300

    @classmethod
    def from_xml(cls, element: NSElement):
        """Parse the XML tag."""
        depth = parse_resolve_depth(element.attrib.get("resolveDepth", None))
        return cls(
            xpath=element.text,
            resolve=ResolveValue[element.attrib.get("resolve", "none")],
            resolve_depth=depth,
            resolve_path=element.attrib.get("resolvePath"),
            resolve_timeout=element.get_int_attribute("resolveTimeout", 300),
            xpath_ns_aliases=element.ns_aliases,
        )

    def parse_xpath(self, feature_types: list) -> XPathMatch:
        """Convert the XPath into the required ORM query elements."""
        if self.resolve_path:
            raise OperationProcessingFailed("resolvePath is not supported", locator="propertyName")
        if "valueOf(" in self.xpath or (self.resolve_path and "valueOf(" in self.resolve_path):
            raise OperationProcessingFailed(
                "valueOf(..) syntax is not supported", locator="propertyName"
            )

        # Can resolve against XSD paths, find the correct DB field name
        return feature_types[0].resolve_element(self.xpath, self.xpath_ns_aliases)

    def decorate_query(self, compiler: CompiledQuery):
        """Update the low-level query based on this property projection."""
        # This will also validate the name because it resolves the ORM path.
        # The actual limiting of fields happens inside the decorate_queryset() of the renderer.
        xpath_match = self.parse_xpath(compiler.feature_types)
        if xpath_match.orm_filters:
            # Apply any xpath [attr=value] lookups.
            compiler.add_extra_lookup(xpath_match.orm_filters)


def parse_resolve_depth(depth: str | None) -> int | None:
    """Parse the resolve depth value."""
    try:
        return None if depth is None or depth == "*" else int(depth)
    except (ValueError, TypeError) as e:
        raise InvalidParameterValue(
            "ResolveDepth must be an integer or '*'", locator="resolveDepth"
        ) from e



"""These classes map to the FES 2.0 specification for operators.
The class names and attributes are identical to those in the FES spec.

Inheritance structure:

* :class:`Operator`

 * :class:`IdOperator` chains ``<fes:ResourceId>``.
 * :class:`NonIdOperator`

  * :class:`ComparisonOperator`

   * :class:`BinaryComparisonOperator` for :class:`BinaryComparisonName` tags like ``<fes:PropertyIsEqualTo>``.
   * :class:`BetweenComparisonOperator` for ``<fes:PropertyIsBetween>``
   * :class:`LikeOperator` for ``<fes:PropertyIsLike>``.
   * :class:`NilOperator` for ``<fes:PropertyIsNil>``.
   * :class:`NullOperator` for ``<fes:PropertyIsNill>``.

  * :class:`SpatialOperator`

   * :class:`DistanceOperator` for the :class:`DistanceOperator` tags: ``<fes:DWithin>`` and ``<fes:Beyond>``.
   * :class:`BinarySpatialOperator` for :class:`SpatialOperatorName` tags like ``<fes:BBOX>``.

  * :class:`TemporalOperator` for :class:`TemporalOperatorName` tags like ``<fes:After>``.
  * :class:`LogicalOperator`

   * :class:`BinaryLogicOperator` for the :class:`BinaryLogicType` tags: ``<fes:And>`` and ``<fes:Or>``.
   * :class:`UnaryLogicOperator` for the :class:`UnaryLogicType` tag: ``<fes:Not>``.

  * :class:`ExtensionOperator` for custom additions.
"""

from __future__ import annotations

import logging
import operator
from dataclasses import dataclass, field
from decimal import Decimal
from enum import Enum
from functools import cached_property, reduce
from itertools import groupby
from typing import ClassVar, Union

from django.contrib.gis import measure
from django.db.models import Q

from gisserver.exceptions import (
    ExternalParsingError,
    InvalidParameterValue,
    OperationProcessingFailed,
)
from gisserver.parsers import gml
from gisserver.parsers.ast import (
    AstNode,
    TagNameEnum,
    expect_children,
    expect_tag,
    tag_registry,
)
from gisserver.parsers.query import CompiledQuery, RhsTypes, ScalarTypes
from gisserver.parsers.values import fix_type_name
from gisserver.parsers.xml import NSElement, xmlns
from gisserver.types import GeometryXsdElement, XPathMatch

from .expressions import Expression, Literal, ValueReference
from .identifiers import Id
from .lookups import ARRAY_LOOKUPS  # also registers the lookups.

logger = logging.getLogger(__name__)

#: Define the types that a ``<gml:SpatialDescription>`` can be:
SpatialDescription = Union[gml.GM_Object, gml.GM_Envelope, ValueReference]

#: Define the types that a ``<gml:TemporalOperand>`` can be:
TemporalOperand = Union[gml.TM_Object, ValueReference]

# Fully qualified tag names
FES_VALUE_REFERENCE = xmlns.fes20.qname("ValueReference")
FES_DISTANCE = xmlns.fes20.qname("Distance")
FES_LOWER_BOUNDARY = xmlns.fes20.qname("LowerBoundary")
FES_UPPER_BOUNDARY = xmlns.fes20.qname("UpperBoundary")
FES1_PROPERTY_NAME = xmlns.fes20.qname("PropertyName")  # old tag sometimes used by clients


class MatchAction(Enum):
    """Values for the 'matchAction' attribute of the :class:`BinaryComparisonOperator`."""

    All = "All"
    Any = "Any"
    One = "One"

    def __repr__(self):
        # Make repr(filter) easier to copy-paste
        return f"{self.__class__.__name__}.{self.name}"


class BinaryComparisonName(TagNameEnum):
    """XML tag names for value comparisons.
    This also maps their names to ORM field lookups.
    """

    PropertyIsEqualTo = "exact"
    PropertyIsNotEqualTo = "fes_notequal"  # custom FesNotEqualTo lookup
    PropertyIsLessThan = "lt"
    PropertyIsGreaterThan = "gt"
    PropertyIsLessThanOrEqualTo = "lte"
    PropertyIsGreaterThanOrEqualTo = "gte"


REVERSE_LOOKUPS = {
    "gt": "lt",
    "gte": "lte",
    "lt": "gt",
    "lte": "gte",
}


class DistanceOperatorName(TagNameEnum):
    """XML tag names mapped to distance operators for the ORM."""

    Beyond = "fes_beyond"  # using __distance_gt=.. would be slower.
    DWithin = "dwithin"  # ST_DWithin uses indices, distance_lte does not.


class SpatialOperatorName(TagNameEnum):
    """XML tag names mapped to geometry operators.

    The values correspond with GeoDjango operators. So a ``BBOX`` query
    will translate into ``geometry__intersects=Polygon(...)``.
    """

    # (A Within B) implies that (B Contains A)

    # BBOX can either be implemented using bboverlaps (more efficient), or the
    # more correct "intersects" option (e.g. a line near the box would match otherwise).
    BBOX = "intersects"  # ISO version: "NOT DISJOINT"
    Equals = "equals"  # Test whether two geometries are topologically equal
    Disjoint = "disjoint"  # Tests whether two geometries are disjoint (do not interact)
    Intersects = "intersects"  # Tests whether two geometries intersect
    Touches = "touches"  # Tests whether two geometries touch (e.g. country border).
    Crosses = "crosses"  # Tests whether two geometries cross (e.g. two streets).
    Within = "within"  # Tests a geometry is within another one (e.g. city within province).
    Contains = "contains"  # Tests a geometry contains another one (e.g. province contains city).
    Overlaps = "overlaps"  # Test whether two geometries overlap


class TemporalOperatorName(TagNameEnum):
    """XML tag names mapped to datetime operators.

    Explanation here: http://old.geotools.org/Temporal-Filters_211091519.html
    and: https://github.com/geotools/geotools/wiki/temporal-filters
    """

    After = "after"
    Before = "before"
    Begins = "begins"
    BegunBy = "begunby"
    TContains = "tcontains"
    TEquals = "tequals"
    TOverlaps = "toverlaps"
    During = "during"
    Meets = "meets"
    OverlappedBy = "overlappedby"
    MetBy = "metby"
    EndedBy = "endedby"
    AnyInteracts = "anyinteracts"


class BinaryLogicType(TagNameEnum):
    """XML tag names for the BinaryLogicOperator."""

    And = operator.and_
    Or = operator.or_


class UnaryLogicType(TagNameEnum):
    """XML tag names for the UnaryLogicOperator."""

    Not = operator.inv


@dataclass
@tag_registry.register("Distance")
class Measure(AstNode):
    """A measurement for a distance element.

    This parses and handles the syntax::

        <fes:Distance uom="...">value</fes:Distance>

    This element is used within the :class:`DistanceOperator` that
    handles the ``<fes:DWithin>`` and ``<fes:Beyond>`` tags.

    The "unit of measurement" (uom) supports most standard units, like meters (``m``),
    kilometers (``km``), nautical mile (``nm``), miles (``mi``), inches (``inch``).
    The full list can be found at: https://docs.djangoproject.com/en/5.1/ref/contrib/gis/measure/#supported-units
    """

    xml_ns = xmlns.fes20

    value: Decimal
    uom: str  # Unit of measurement, fes20:UomSymbol | fes20:UomURI

    @classmethod
    @expect_tag(xmlns.fes20, "Distance")
    def from_xml(cls, element: NSElement):
        return cls(value=Decimal(element.text), uom=element.get_str_attribute("uom"))

    def build_rhs(self, compiler) -> measure.Distance:
        return measure.Distance(default_unit=self.uom, **{self.uom: self.value})


class Operator(AstNode):
    """Abstract base class, as defined by FES spec.

    This base class is also used in parsing; for example the ``<fes:Filter>``
    tag only allows ``Operator`` and ``Expression`` subclasses as allowed arguments.
    Having all those classes as Python types, makes it very easy to validate
    whether a given child element is the expected node type.
    """

    xml_ns = xmlns.fes20

    def build_query(self, compiler: CompiledQuery) -> Q | None:
        raise NotImplementedError(f"Using {self.__class__.__name__} is not supported yet.")


@dataclass
class IdOperator(Operator):
    """List of :class:`~gisserver.parsers.fes20.identifers.ResourceId` objects.

    A ``<fes:Filter>`` only has a single predicate.
    Hence, this operator is used to wrap the ``<fes:ResourceId>`` elements in the syntax::

        <fes:Filter>
            <fes:ResourceId rid="typename.123" />
            <fes:ResourceId rid="typename.345" />
        </fes:Filter>
    """

    id: list[Id]

    def get_type_names(self) -> list[str]:
        """Provide a list of all type names accessed by this operator"""
        return [type_name for type_name in self.grouped_ids if type_name is not None]

    @cached_property
    def grouped_ids(self) -> dict[str, list[Id]]:
        # For itertools.groupby(), items have to be sorted first.
        ids = sorted(self.id, key=operator.attrgetter("rid"))
        return {
            type_name: list(items)
            for type_name, items in groupby(ids, key=lambda id: id.get_type_name())
        }

    def build_query(self, compiler):
        """Generate the ID lookup query.

        As these identifiers also reference the type name, no Q-object is
        returned. The lookups are directly added to the fes query object.
        """
        # When ResourceId has no type_name associated, this means that the
        # parsing was invalid. Instead of raising an error, empty results are returned.
        if len(self.grouped_ids) == 1 and self.grouped_ids.get(None):
            compiler.mark_empty()
            return

        type_names = {ft.xml_name for ft in compiler.feature_types}
        for type_name, items in self.grouped_ids.items():
            type_name = fix_type_name(type_name, compiler.feature_types[0].xml_namespace)
            if type_name not in type_names:
                # When ResourceId + typenames is defined, it should be a value from typenames see WFS spec 7.9.2.4.1
                # This is tested here for RESOURCEID with a typename.id format.
                # Otherwise, this breaks the CITE RESOURCEID=test-UUID parameter.
                raise InvalidParameterValue(
                    "When TYPENAMES and RESOURCEID are combined, "
                    "the RESOURCEID type should be included in TYPENAMES.",
                    locator="resourceId",
                )

            ids_subset = reduce(operator.or_, [id.build_query(compiler) for id in items])
            compiler.add_lookups(ids_subset, type_name=type_name)


class NonIdOperator(Operator):
    """Abstract base class, as defined by FES spec.

    This is used for nearly all operators,
    except those that have ``<fes:ResourceId>`` elements as children.

    Some operators, such as the ``<fes:And>``, ``<fes:Or>`` and ``<fes:Not>`` operators
    explicitly support only ``NonIdOperator`` elements as arguments.
    Hence, having this base class as Python type simplifies parsing.
    """

    _source = None  # often declared as non-comparing in dataclasses below.
    allow_geometries = False

    def build_compare(
        self,
        compiler: CompiledQuery,
        lhs: Expression,
        lookup: str,
        rhs: Expression | RhsTypes,
    ) -> Q:
        """Use the value in comparison with some other expression.

        This calls build_lhs() and build_rhs() on the expressions.
        """
        # lhs and rhs are allowed to be reversed. However, the SQL compiler
        # works much simpler when Django can predict the actual data type.
        if isinstance(lhs, Literal) and isinstance(rhs, ValueReference):
            logger.debug("Filter switches lhs/rhs for %s %s %s", lhs.raw_value, lookup, rhs.xpath)
            lhs, rhs = rhs, lhs
            lookup = REVERSE_LOOKUPS.get(lookup, lookup)  # >= should become <=

        lookup = self.validate_comparison(compiler, lhs, lookup, rhs)

        # Build Django Q-object
        lhs_orm_name = lhs.build_lhs(compiler)

        if isinstance(rhs, (Expression, gml.GM_Object)):
            rhs = rhs.build_rhs(compiler)
            if (
                isinstance(lhs, ValueReference)
                and isinstance(rhs, ScalarTypes)
                and lookup != "isnull"
            ):
                # Building the expression resolved as a scalar after all (for BinaryOperator).
                # It allows performing a better type check:
                xsd_element = lhs.parse_xpath(compiler.feature_types).child
                xsd_element.to_python(rhs)

        comparison = Q(**{f"{lhs_orm_name}__{lookup}": rhs})
        return compiler.apply_extra_lookups(comparison)

    def validate_comparison(
        self,
        compiler: CompiledQuery,
        lhs: Expression,
        lookup: str,
        rhs: Expression | gml.GM_Object | RhsTypes,
    ) -> str:
        """Validate whether a given comparison is even possible.

        For example, comparisons like ``name == "test"`` are fine,
        but ``geometry < 4`` or ``datefield == 35.2`` raise an error.

        :param compiler: The object that holds the intermediate state
        :param lhs: The left-hand-side of the comparison (e.g. the element).
        :param lookup: The ORM lookup expression being used (e.g. ``equals`` or ``fes_like``).
        :param rhs: The right-hand-side of the comparison (e.g. the value).
        """
        if isinstance(lhs, Literal) and isinstance(rhs, ValueReference):
            lhs, rhs = rhs, lhs

        if isinstance(lhs, ValueReference):
            xsd_element = lhs.parse_xpath(compiler.feature_types).child
            tag = self._source if self._source is not None else None

            # e.g. deny <PropertyIsLessThanOrEqualTo> against <gml:boundedBy>
            if xsd_element.type.is_geometry and not self.allow_geometries:
                raise OperationProcessingFailed(
                    f"Operator '{tag}' does not support comparing"
                    f" geometry properties: '{xsd_element.xml_name}'.",
                    locator="filter",
                    status_code=400,  # not HTTP 500 here. Spec allows both.
                )

            # Checking scalar values against array fields will fail.
            # However, to make the queries consistent with other unbounded types (i.e. M2M fields),
            # it makes sense to return an object when *one* entry in the array matches.
            if xsd_element.is_array:
                try:
                    lookup = ARRAY_LOOKUPS[lookup]
                except KeyError:
                    logger.debug("No array lookup known for %s", lookup)
                    raise OperationProcessingFailed(
                        f"Operator '{tag}' is not supported for "
                        f"the '{xsd_element.name}' property.",
                        locator="filter",
                        status_code=400,  # not HTTP 500 here. Spec allows both.
                    ) from None

            if isinstance(rhs, Literal):
                # Since the element is resolved, inform the Literal how to parse the value.
                # This avoids various validation errors along the path.
                rhs.bind_type(xsd_element.type)

                # When a common case of value comparison is done, the inputs
                # can be validated before the ORM query is constructed.
                xsd_element.validate_comparison(rhs.raw_value, lookup=lookup, tag=tag)
            elif isinstance(rhs, ScalarTypes) and lookup != "isnull":
                # Can still compare two elements, e.g. date >= ((2020 - 12) - 10) by QGis
                xsd_element.validate_comparison(rhs, lookup=lookup, tag=tag)

        return lookup

    def build_compare_between(
        self,
        compiler: CompiledQuery,
        lhs: Expression,
        lookup: str,
        rhs: tuple[Expression | ValueReference | gml.GM_Object, Expression | Measure],
    ) -> Q:
        """Use the value in comparison with 2 other values (e.g. between query)"""
        self.validate_comparison(compiler, lhs, lookup, rhs[0])
        self.validate_comparison(compiler, lhs, lookup, rhs[1])

        field_name = lhs.build_lhs(compiler)
        comparison = Q(
            **{
                f"{field_name}__{lookup}": (
                    rhs[0].build_rhs(compiler),
                    rhs[1].build_rhs(compiler),
                )
            }
        )
        return compiler.apply_extra_lookups(comparison)


class SpatialOperator(NonIdOperator):
    """Abstract base class, as defined by FES spec."""


@dataclass
@tag_registry.register("Beyond")
@tag_registry.register("DWithin")
class DistanceOperator(SpatialOperator):
    """Comparing the distance to a geometry.

    This parses and handles the syntax::

        <fes:DWithin>
            <fes:ValueReference>geometry</fes:ValueReference>
            <gml:Point srsDimension="2">
                <gml:pos>43.55749 1.525864</gml:pos>
            </gml:Point>
            <fes:Distance oum="m:>100</fes:Distance>
        </fes:DWithin>
    """

    allow_geometries = True  # override static attribute

    valueReference: ValueReference
    operatorType: DistanceOperatorName
    geometry: gml.GM_Object
    distance: Measure
    _source: str | None = field(compare=False, default=None, repr=False)

    @classmethod
    @expect_children(3, ValueReference, gml.GM_Object, Measure)
    def from_xml(cls, element: NSElement):
        geometries = gml.find_gml_nodes(element)
        if not geometries:
            raise ExternalParsingError(f"Missing gml element in <{element.tag}>")
        elif len(geometries) > 1:
            raise ExternalParsingError(f"Multiple gml elements found in <{element.tag}>")

        return cls(
            valueReference=ValueReference.from_xml(element.find(FES_VALUE_REFERENCE)),
            operatorType=DistanceOperatorName.from_xml(element),
            geometry=gml.parse_gml_node(geometries[0]),
            distance=Measure.from_xml(element.find(FES_DISTANCE)),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        return self.build_compare_between(
            compiler,
            lhs=self.valueReference,
            lookup=self.operatorType.value,
            rhs=(self.geometry, self.distance),
        )


@dataclass
@tag_registry.register(SpatialOperatorName)  # <BBOX>, <Equals>, ...
class BinarySpatialOperator(SpatialOperator):
    """A comparison of geometries using 2 values, e.g. A Within B.

    This parses and handles the syntax, and its variants::

        <fes:BBOX>
            <fes:ValueReference>Geometry</fes:ValueReference>
            <gml:Envelope srsName="http://www.opengis.net/def/crs/epsg/0/4326">
                <gml:lowerCorner>13.0983 31.5899</gml:lowerCorner>
                <gml:upperCorner>35.5472 42.8143</gml:upperCorner>
            </gml:Envelope>
        </fes:BBOX>

    It also handles the ``<fes:Equals>``, ``<fes:Within>``, ``<fes:Intersects>``, etc..
    that exist in the :class:`SpatialOperatorName` enum.
    """

    allow_geometries = True  # override static attribute

    operatorType: SpatialOperatorName
    operand1: ValueReference | None
    operand2: SpatialDescription
    _source: str | None = field(compare=False, default=None, repr=False)

    @classmethod
    def from_xml(cls, element: NSElement):
        operator_type = SpatialOperatorName.from_xml(element)
        if operator_type is SpatialOperatorName.BBOX and len(element) == 1:
            # For BBOX, the geometry operator is optional
            ref = None
            geo = element[0]
        else:
            if len(element) != 2:
                raise ExternalParsingError(f"{element.tag} should have 2 operators")
            ref, geo = list(element)

        return cls(
            operatorType=operator_type,
            operand1=ValueReference.from_xml(ref) if ref is not None else None,
            operand2=tag_registry.node_from_xml(geo, allowed_types=SpatialDescription.__args__),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        operant1 = self.operand1
        if operant1 is None:
            # BBOX element points to the existing geometry element by default.
            operant1 = _ResolvedValueReference(compiler.feature_types[0].main_geometry_element)

        return self.build_compare(
            compiler,
            lhs=operant1,
            lookup=self.operatorType.value,
            rhs=self.operand2,
        )


class _ResolvedValueReference(ValueReference):
    """Internal ValueReference subclass that already knows which element it resolve to.
    This avoids translating back and forth between an XPath path and the desired node.
    """

    def __init__(self, geo_element: GeometryXsdElement):
        self.xpath = f"(mocked {geo_element.orm_path})"  # keep __str__ and __repr__ happy.
        self.xpath_ns_aliases = {}
        self.geo_element = geo_element
        self.orm_path = geo_element.orm_path

    def parse_xpath(
        self, feature_types: list, ns_aliases: dict[str, str] | None = None
    ) -> XPathMatch:
        return XPathMatch(feature_types[0], [self.geo_element], query="")

    def build_lhs(self, compiler: CompiledQuery):
        return self.orm_path


@dataclass
@tag_registry.register(TemporalOperatorName, hidden=True)  # <After>, <Before>, ...
class TemporalOperator(NonIdOperator):
    """Comparisons with dates.

    For these operators, only the parsing is implemented.
    These are not translated into ORM queries yet.

    It supports a syntax such as::

       <fes:TEquals>
          <fes:ValueReference>SimpleTrajectory/gml:validTime/gml:TimeInstant</fes:ValueReference>
          <gml:TimeInstant gml:id="TI1">
             <gml:timePosition>2005-05-19T09:28:40Z</gml:timePosition>
          </gml:TimeInstant>
       </fes:TEquals>

    or::

       <fes:During>
          <fes:ValueReference>SimpleTrajectory/gml:validTime/gml:TimeInstant</fes:ValueReference>
          <gml:TimePeriod gml:id="TP1">
             <gml:begin>
                <gml:TimeInstant gml:id="TI1">
                   <gml:timePosition>2005-05-17T00:00:00Z</gml:timePosition>
                </gml:TimeInstant>
             </gml:begin>
             <gml:end>
                <gml:TimeInstant gml:id="TI2">
                   <gml:timePosition>2005-05-23T00:00:00Z</gml:timePosition>
                </gml:TimeInstant>
             </gml:end>
          </gml:TimePeriod>
       </fes:During>
    """

    operatorType: TemporalOperatorName
    operand1: ValueReference
    operand2: TemporalOperand
    _source: str | None = field(compare=False, default=None, repr=False)

    @classmethod
    @expect_children(2, ValueReference, *TemporalOperand.__args__)
    def from_xml(cls, element: NSElement):
        return cls(
            operatorType=TemporalOperatorName.from_xml(element),
            operand1=ValueReference.from_xml(element[0]),
            operand2=tag_registry.node_from_xml(
                element[1], allowed_types=TemporalOperand.__args__
            ),
            _source=element.tag,
        )


class ComparisonOperator(NonIdOperator):
    """Base class for comparisons.
    This class name mirrors the fes-spec name,
    and allows grouping various comparisons together.
    """


@dataclass
@tag_registry.register(BinaryComparisonName)  # <PropertyIs...>
class BinaryComparisonOperator(ComparisonOperator):
    """A comparison between 2 values, e.g. *A == B*.

    This parses and handles the syntax::

        <fes:PropertyIsEqualTo>
            <fes:ValueReference>city/name</fes:ValueReference>
            <fes:Literal>CloudCity</fes:Literal>
        </fes:PropertyIsEqualTo>

    ...and all variations (``<fes:PropertyIsLessThan>``, etc...)
    that are listed in the :class:`BinaryComparisonName` enum.

    Note that both arguments are expressions, so these can be value references, literals
    or functions. The ``<fes:Literal>`` element may hold a GML element as its value.
    """

    operatorType: BinaryComparisonName
    expression: tuple[Expression, Expression]
    matchCase: bool = True
    matchAction: MatchAction = MatchAction.Any
    _source: str | None = field(compare=False, default=None)

    @classmethod
    @expect_children(2, Expression, silent_allowed=(FES1_PROPERTY_NAME,))
    def from_xml(cls, element: NSElement):
        return cls(
            operatorType=BinaryComparisonName.from_xml(element),
            expression=(
                Expression.child_from_xml(element[0]),
                Expression.child_from_xml(element[1]),
            ),
            matchCase=element.get("matchCase", True),
            matchAction=MatchAction(element.get("matchAction", default=MatchAction.Any)),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        lhs, rhs = self.expression
        return self.build_compare(compiler, lhs=lhs, lookup=self.operatorType.value, rhs=rhs)


@dataclass
@tag_registry.register("PropertyIsBetween")
class BetweenComparisonOperator(ComparisonOperator):
    """Check whether a value is between two elements.

    This parses and handles the syntax::

        <fes:PropertyIsBetween>
            <fes:ValueReference>DEPTH</fes:ValueReference>
            <fes:LowerBoundary><fes:Literal>100</fes:Literal></fes:LowerBoundary>
            <fes:UpperBoundary><fes:Literal>200</fes:Literal></fes:UpperBoundary>
        </fes:PropertyIsBetween>

    Note that both boundary arguments receive expressions, so these can be value
    references, literals or functions!
    """

    expression: Expression
    lowerBoundary: Expression
    upperBoundary: Expression
    _source: str | None = field(compare=False, default=None, repr=False)

    @classmethod
    @expect_children(3, Expression, FES_LOWER_BOUNDARY, FES_UPPER_BOUNDARY)
    def from_xml(cls, element: NSElement):
        if (element[1].tag != FES_LOWER_BOUNDARY) or (element[2].tag != FES_UPPER_BOUNDARY):
            raise ExternalParsingError(
                f"{element.tag} should have 3 child nodes: "
                f"(expression), <LowerBoundary>, <UpperBoundary>"
            )

        lower = element[1]
        upper = element[2]

        if len(lower) != 1:
            raise ExternalParsingError(f"{lower.tag} should have 1 expression child node")
        if len(upper) != 1:
            raise ExternalParsingError(f"{upper.tag} should have 1 expression child node")

        return cls(
            expression=Expression.child_from_xml(element[0]),
            lowerBoundary=Expression.child_from_xml(lower[0]),
            upperBoundary=Expression.child_from_xml(upper[0]),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        return self.build_compare_between(
            compiler,
            lhs=self.expression,
            lookup="range",
            rhs=(self.lowerBoundary, self.upperBoundary),
        )


@dataclass
@tag_registry.register("PropertyIsLike")
class LikeOperator(ComparisonOperator):
    """Perform wildcard matching.

    This parses and handles the syntax::

        <fes:PropertyIsLike wildCard="*" singleChar="#" escapeChar="!">
            <fes:ValueReference>last_name</fes:ValueReference>
            <fes:Literal>John*</fes:Literal>
        </fes:PropertyIsLike>
    """

    expression: tuple[Expression, Expression]
    wildCard: str
    singleChar: str
    escapeChar: str
    _source: str | None = field(compare=False, default=None, repr=False)

    @classmethod
    @expect_children(2, Expression)
    def from_xml(cls, element: NSElement):
        return cls(
            expression=(
                Expression.child_from_xml(element[0]),
                Expression.child_from_xml(element[1]),
            ),
            # These attributes are required by the WFS spec:
            wildCard=element.get_str_attribute("wildCard"),
            singleChar=element.get_str_attribute("singleChar"),
            escapeChar=element.get_str_attribute("escapeChar"),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        lhs, rhs = self.expression
        if isinstance(rhs, Literal):
            value = str(rhs.value)  # value could be auto-casted to int.

            # Not using r"\" here as that is a syntax error.
            if self.escapeChar != "\\":
                value = value.replace("\\", "\\\\").replace(self.escapeChar, "\\")
            if self.wildCard != "%":
                value = value.replace("%", r"\%").replace(self.wildCard, "%")
            if self.singleChar != "_":
                value = value.replace("_", r"\_").replace(self.singleChar, "_")

            rhs = Literal(raw_value=value)
        else:
            raise ExternalParsingError(
                f"Expected a literal value for the {self.xml_name} operator."
            )

        # Use the FesLike lookup
        return self.build_compare(compiler, lhs=lhs, lookup="fes_like", rhs=rhs)


@dataclass
@tag_registry.register("PropertyIsNil")
class NilOperator(ComparisonOperator):
    """Check whether the value evaluates to null/None.
    If the WFS returned a property element with ``<app:field xsi:nil='true'>``, this returns true.

    It parses and handles syntax such as::

        <fes:PropertyIsNil>
            <fes:ValueReference>city/name</fes:ValueReference>
        </fes:PropertyIsNil>

    Note that the provided argument can be any expression, not just a value reference.
    Thus, it can also check whether a function returns null.
    """

    expression: Expression | None
    nilReason: str
    _source: str | None = field(compare=False, default=None, repr=False)

    # Allow checking whether the geometry is null
    allow_geometries: ClassVar[bool] = True

    @classmethod
    @expect_children(1, Expression)
    def from_xml(cls, element: NSElement):
        return cls(
            expression=Expression.child_from_xml(element[0]) if element is not None else None,
            nilReason=element.get("nilReason"),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        # Any value that evaluates to None is returned as 'xs:nil' in our output.
        return self.build_compare(compiler, lhs=self.expression, lookup="isnull", rhs=True)


@dataclass
@tag_registry.register("PropertyIsNull")
class NullOperator(ComparisonOperator):
    """Check whether the property exists.
    If the WFS would not return the property element ``<app:field>``, this returns true.

    It parses and handles syntax such as::

        <fes:PropertyIsNull>
            <fes:ValueReference>city/name</fes:ValueReference>
        </fes:PropertyIsNull>

    Note that the provided argument can be any expression, not just a value reference.
    """

    expression: Expression
    _source: str | None = field(compare=False, default=None, repr=False)

    # Allow checking whether the geometry is null
    allow_geometries: ClassVar[bool] = True

    @classmethod
    @expect_children(1, Expression)
    def from_xml(cls, element: NSElement):
        return cls(expression=Expression.child_from_xml(element[0]), _source=element.tag)

    def build_query(self, compiler: CompiledQuery) -> Q:
        # For now, the implementation is identical to PropertyIsNil.
        # According to the WFS spec, this should only be true when the element
        # is not returned at all (minOccurs=0).
        # TODO: this happens for maxOccurs=unbounded with a null value.
        return self.build_compare(compiler, lhs=self.expression, lookup="isnull", rhs=True)


class LogicalOperator(NonIdOperator):
    """Base class in the fes-spec for AND, OR, NOT comparisons"""


@dataclass
@tag_registry.register("And")
@tag_registry.register("Or")
class BinaryLogicOperator(LogicalOperator):
    """Apply an 'AND' or 'OR' operator.

    This parses and handles the syntax::

        <fes:And>
            <fes:PropertyIsGreaterThanOrEqualTo>
                ...
            </fes:PropertyIsGreaterThanOrEqualTo>
            <fes:BBOX>
                ...
            </fes:BBOX>
        </fes:And>

    Any tag deriving from :class:`NonIdOperator` is allowed here.
    """

    operands: list[NonIdOperator]
    operatorType: BinaryLogicType
    _source: str | None = field(compare=False, default=None, repr=False)

    @classmethod
    @expect_children(2, NonIdOperator)
    def from_xml(cls, element: NSElement):
        return cls(
            operands=[NonIdOperator.child_from_xml(child) for child in element],
            operatorType=BinaryLogicType.from_xml(element),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        """Apply the AND/OR operation to the Q-object"""
        values = [q.build_query(compiler) for q in self.operands]
        return reduce(self.operatorType.value, values)


@dataclass
@tag_registry.register("Not")
class UnaryLogicOperator(LogicalOperator):
    """Apply a NOT operator.

    This parses and handles the syntax::

        <fes:Not>
            <fes:PropertyIsNil>
                <fes:ValueReference>city/name</fes:ValueReference>
            </fes:PropertyIsNil>
        </fes:Not>
    """

    operands: NonIdOperator
    operatorType: UnaryLogicType
    _source: str | None = field(compare=False, default=None, repr=False)

    @classmethod
    @expect_children(1, NonIdOperator)
    def from_xml(cls, element: NSElement):
        return cls(
            operands=NonIdOperator.child_from_xml(element[0]),
            operatorType=UnaryLogicType.from_xml(element),
            _source=element.tag,
        )

    def build_query(self, compiler: CompiledQuery) -> Q:
        """Apply the NOT operation to the Q-object"""
        return self.operatorType.value(self.operands.build_query(compiler))


class ExtensionOperator(NonIdOperator):
    """Base class for extensions to FES 2.0.

    It's fully allowed to introduce new operators on your own namespace.
    These need to inherit from this class.
    """



"""Generic Open Web Services (OWS) protocol bits to handle incoming requests.

This is the common logic between WFS, WMS and other protocols.

These translate both request-syntax formats in the same internal objects
that the rest of the controller/view logic can use.
"""

from .kvp import KVPRequest
from .requests import (
    BaseOwsRequest,
    parse_get_request,
    parse_post_request,
    resolve_kvp_parser_class,
    resolve_xml_parser_class,
)

__all__ = (
    "KVPRequest",
    "BaseOwsRequest",
    "resolve_kvp_parser_class",
    "resolve_xml_parser_class",
    "parse_get_request",
    "parse_post_request",
)



"""Functions to be callable from query filters.

By using the :attr:`function_registry`, custom stored functions can be registered in this server.
These are called by the filter queries using the :class:`~gisserver.parsers.fes20.expressions.Function` element.
Out of the box, various built-in functions are present.

Built-in options are documented in :ref:`functions`.

Most of the out-of-the box options are inspired
by `GeoServer <https://docs.geoserver.org/latest/en/user/filter/function_reference.html>`_.
Functions which already have a fes-syntax equivalent have been are omitted.
"""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Union

import django
from django.contrib.gis.db.models import functions as gis
from django.db import models
from django.db.models import functions
from django.db.models.expressions import Combinable, Value

from gisserver.exceptions import InvalidParameterValue
from gisserver.types import XsdTypes

__all__ = ["function_registry", "FesFunctionRegistry", "FesFunction"]

FesArg = Union[Combinable, models.Q]
FesFunctionBody = Union[models.Func, Callable[..., models.Func]]


@dataclass(order=True)
class FesFunction:
    """A registered database function that can be used by ``<fes:Function name="...">``.

    The :class:`~gisserver.parsers.fes20.expressions.Function` class will resolve
    these registered functions by name, and call :meth:`build_query` to include them
    in the database query. This will actually insert a Django ORM function in the query!

    This wrapper class also provides the metadata and type descriptions of the function,
    which is exposed in the ``GetCapabilities`` call.
    """

    #: Name of the function
    name: str

    #: The function body
    body: FesFunctionBody

    #: The argument names with their XSD types
    arguments: dict[str, XsdTypes] | None = None

    #: The XSD return type
    returns: XsdTypes | None = None

    def build_query(self, *expressions: FesArg) -> models.Func:
        """Build the query expression for the function."""
        if len(expressions) != len(self.arguments):
            # Avoid passing extra parameters to the function if those are not defined.
            raise TypeError(f'Invalid number of arguments for <fes:Function name="{self.name}">')

        # Keyword arguments are avoided, since some Django SQL functions had
        # SQL-injection issues with those arguments. Only passing expressions for now.
        return self.body(*expressions)


class FesFunctionRegistry:
    """Registry of functions to be callable by ``<fes:Function>``.

    The registered functions should be capable of running an SQL function.
    """

    def __init__(self):
        self.functions = {}

    def __bool__(self):
        """Tell whether there are functions"""
        return bool(self.functions)

    def __iter__(self):
        """Iterate over the functions"""
        return iter(sorted(self.functions.values()))  # for template rendering

    def register(
        self,
        name=None,
        body: FesFunctionBody | None = None,
        *,
        arguments: dict[str, XsdTypes] | None = None,
        returns: XsdTypes | None = None,
    ):
        """Decorator to register a function.

        It's not recommended to register the Django Func objects directly,
        as the parameters are passed on from the client-side. Instead, create
        a wrapper function that enforces a controlled set of parameters.
        """

        def _wrapper(callable: FesFunctionBody):
            function = FesFunction(
                name=name or callable.__name__,
                body=callable,
                arguments=arguments,
                returns=returns,
            )
            self.functions[function.name] = function
            return callable

        if body is not None:
            # Usage as direct registration
            return _wrapper(body)
        else:
            # Usage as decorator
            return _wrapper

    def resolve_function(self, function_name) -> FesFunction:
        """Resole the function using its name."""
        try:
            return self.functions[function_name]
        except KeyError:
            raise InvalidParameterValue(
                f"Unsupported function: {function_name}", locator="filter"
            ) from None


#: The function registry
function_registry = FesFunctionRegistry()


# Register a set of default SQL functions.

# -- strings

function_registry.register(
    "strConcat",
    functions.Concat,
    arguments={"string1": XsdTypes.string, "string2": XsdTypes.string},
    returns=XsdTypes.string,
)

function_registry.register(
    "strIndexOf",
    lambda string, substring: (functions.StrIndex(string, Value(substring)) - 1),
    arguments={"string": XsdTypes.string, "substring": XsdTypes.string},
    returns=XsdTypes.string,
)

function_registry.register(
    "strSubstring",
    lambda string, begin, end: functions.Substr(string, begin + 1, end - begin),
    arguments={"string": XsdTypes.string, "begin": XsdTypes.integer, "end": XsdTypes.integer},
    returns=XsdTypes.string,
)

function_registry.register(
    "strSubstringStart",
    lambda string, begin, end: functions.Substr(string, begin + 1),
    arguments={"string": XsdTypes.string, "begin": XsdTypes.integer},
    returns=XsdTypes.string,
)

function_registry.register(
    "strToLowerCase",
    functions.Lower,
    arguments={"string": XsdTypes.string},
    returns=XsdTypes.string,
)

function_registry.register(
    "strToUpperCase",
    functions.Upper,
    arguments={"string": XsdTypes.string},
    returns=XsdTypes.string,
)

function_registry.register(
    "strTrim",
    functions.Trim,
    arguments={"string": XsdTypes.string},
    returns=XsdTypes.string,
)

function_registry.register(
    "strLength",
    functions.Length,
    arguments={"string": XsdTypes.string},
    returns=XsdTypes.integer,
)

function_registry.register(
    "length",
    functions.Length,
    arguments={"string": XsdTypes.string},
    returns=XsdTypes.integer,
)

# -- math numbers

function_registry.register(
    "abs",
    functions.Abs,
    arguments={"number": XsdTypes.decimal},
    returns=XsdTypes.double,
)

function_registry.register(
    "ceil",
    functions.Ceil,
    arguments={"number": XsdTypes.decimal},
    returns=XsdTypes.decimal,
)

function_registry.register(
    "floor",
    functions.Floor,
    arguments={"number": XsdTypes.decimal},
    returns=XsdTypes.decimal,
)

function_registry.register(
    "min",
    functions.Least,
    arguments={"value1": XsdTypes.double, "value2": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "max",
    functions.Greatest,
    arguments={"value1": XsdTypes.double, "value2": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "pow",
    functions.Power,
    arguments={"base": XsdTypes.double, "exponent": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "round",
    functions.Round,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.integer,
)

function_registry.register(
    "exp",
    functions.Exp,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "log",
    functions.Log,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "sqrt",
    functions.Sqrt,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

# -- math trigonometry

function_registry.register(
    "acos",
    functions.ACos,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "asin",
    functions.ASin,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "atan",
    functions.ATan,
    arguments={"value": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "atan2",
    functions.ATan2,
    arguments={"x": XsdTypes.double, "y": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "cos",
    functions.Cos,
    arguments={"radians": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "sin",
    functions.Sin,
    arguments={"radians": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "tan",
    functions.Tan,
    arguments={"radians": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "pi",
    functions.Pi,
    returns=XsdTypes.double,
)

function_registry.register(
    "toDegrees",
    functions.Degrees,
    arguments={"radians": XsdTypes.double},
    returns=XsdTypes.double,
)

function_registry.register(
    "toRadians",
    functions.Radians,
    arguments={"degree": XsdTypes.double},
    returns=XsdTypes.double,
)

# -- geometric

function_registry.register(
    "area",
    gis.Area,
    arguments={"geom": XsdTypes.gmlAbstractGeometryType},
    returns=XsdTypes.double,
)

function_registry.register(
    "centroid",
    gis.Centroid,
    arguments={"geom": XsdTypes.gmlAbstractGeometryType},
    returns=XsdTypes.string,
)

function_registry.register(
    "difference",
    gis.Difference,
    arguments={
        "a": XsdTypes.gmlAbstractGeometryType,
        "b": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.string,
)

function_registry.register(
    "distance",
    gis.Distance,
    arguments={
        "a": XsdTypes.gmlAbstractGeometryType,
        "b": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.double,
)

function_registry.register(
    "envelope",
    gis.Envelope,
    arguments={"geom": XsdTypes.gmlAbstractGeometryType},
    returns=XsdTypes.gmlAbstractGeometryType,  # returns point or polygon
)

function_registry.register(
    "geomLength",
    gis.Length,
    arguments={"geometry": XsdTypes.gmlAbstractGeometryType},
    returns=XsdTypes.float,
)

function_registry.register(
    "intersection",
    gis.Intersection,
    arguments={
        "a": XsdTypes.gmlAbstractGeometryType,
        "b": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.gmlAbstractGeometryType,
)

if django.VERSION >= (4, 2):
    function_registry.register(
        "isEmpty",
        gis.IsEmpty,
        arguments={
            "geom": XsdTypes.gmlAbstractGeometryType,
        },
        returns=XsdTypes.boolean,
    )

function_registry.register(
    "isValid",
    gis.IsValid,
    arguments={
        "geom": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.boolean,
)

function_registry.register(
    "numGeometries",
    gis.NumGeometries,
    arguments={
        "collection": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.integer,
)

function_registry.register(
    "numPoints",
    gis.NumPoints,
    arguments={
        "collection": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.integer,
)

function_registry.register(
    "perimeter",
    gis.Perimeter,
    arguments={
        "geom": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.integer,
)

function_registry.register(
    "symDifference",
    gis.SymDifference,
    arguments={
        "a": XsdTypes.gmlAbstractGeometryType,
        "b": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.gmlAbstractGeometryType,
)

function_registry.register(
    "union",
    gis.Union,
    arguments={
        "a": XsdTypes.gmlAbstractGeometryType,
        "b": XsdTypes.gmlAbstractGeometryType,
    },
    returns=XsdTypes.gmlAbstractGeometryType,
)



"""Output rendering logic for GeoJSON."""

from datetime import datetime, timezone
from decimal import Decimal
from io import BytesIO

import orjson
from django.contrib.gis.db.models.functions import AsGeoJSON
from django.contrib.gis.gdal import AxisOrder
from django.db import models
from django.utils.functional import Promise

from gisserver import conf
from gisserver.crs import CRS84, WGS84
from gisserver.db import get_db_geometry_target
from gisserver.projection import FeatureProjection
from gisserver.types import XsdElement

from .base import CollectionOutputRenderer


def _json_default(obj):
    """Serialize non-built in values to JSON"""
    if isinstance(obj, (Decimal, Promise)):
        return str(obj)
    raise TypeError(f"Unable to serialize {obj.__class__.__name__} to JSON")


class GeoJsonRenderer(CollectionOutputRenderer):
    """Fast GeoJSON renderer, using a stream response.

    The complex encoding bits are handled by the C-library "orjson"
    and the geojson property of GEOSGeometry.

    While Django has a GeoJSON serializer
    (see https://docs.djangoproject.com/en/3.0/ref/contrib/gis/serializers/),
    it does not offer streaming response handling.
    """

    content_type = "application/geo+json; charset=utf-8"
    content_disposition = 'inline; filename="{typenames} {page} {date}.geojson"'
    max_page_size = conf.GISSERVER_GEOJSON_MAX_PAGE_SIZE
    chunk_size = 40_000

    def decorate_queryset(
        self,
        projection: FeatureProjection,
        queryset: models.QuerySet,
    ):
        """Redefine which fields to query, always include geometry, but remove all others"""
        # make sure output CRS matches the coordinate ordering that GEOSGeometry.json returns
        if projection.output_crs == WGS84:
            projection.output_crs = CRS84

        # Make sure geometry is always queried.
        # Other geometries can be excluded as these are not rendered by 'properties'
        main_geo_element = projection.feature_type.main_geometry_element
        projection.add_field(main_geo_element)
        projection.remove_fields(
            lambda element: element.type.is_geometry and element is not main_geo_element
        )

        # Apply the normal optimizations with this altered projection
        return super().decorate_queryset(projection, queryset)

    def render_stream(self):
        self.output = output = BytesIO()

        # Generate the header from a Python dict,
        # but replace the last "}" into a comma, to allow writing more
        header = self.get_header()

        # Have a temporary buffer that is written
        output.write(orjson.dumps(header)[:-1])
        output.write(b',\n  "features": [\n')

        # Flatten the results, they are not grouped in a second FeatureCollection
        is_first_collection = True
        for sub_collection in self.collection.results:
            projection = sub_collection.projection

            if is_first_collection:
                is_first_collection = False
            else:
                output.write(b",\n")

            is_first = True
            for instance in self.read_features(sub_collection):
                if is_first:
                    is_first = False
                else:
                    output.write(b",\n")

                # The "properties" object is generated by orjson.dumps(),
                # while the "geometry" object uses the built-in 'GEOSGeometry.json' result.
                output.write(self.render_feature(projection, instance))

                # Only perform a 'yield' every once in a while,
                # as it goes back-and-forth for writing it to the client.
                if output.tell() > self.chunk_size:
                    json_chunk = output.getvalue()
                    output.seek(0)
                    output.truncate(0)
                    yield json_chunk

        # Instead of performing an expensive .count() on the start of the page,
        # write this as a last field at the end of the response.
        # This still honors the WFS 30 DRAFT without sacrificing performance.
        output.write(b"\n  ],\n")
        footer = self.get_footer()
        output.write(orjson.dumps(footer)[1:])
        output.write(b"\n")
        yield output.getvalue()

    def render_exception(self, exception: Exception):
        """Render the exception in a format that fits with the output."""
        message = super().render_exception(exception)
        buffer = self.output.getvalue().decode()
        return f"{buffer}/* {message} */\n"

    def render_feature(self, projection: FeatureProjection, instance: models.Model) -> bytes:
        """Render the output of a single feature"""

        # Get all instance attributes:
        properties = self.get_properties(projection, projection.xsd_root_elements, instance)
        feature_type = projection.feature_type

        # Add the name field
        if feature_type.show_name_field:
            name = feature_type.get_display_value(instance)
            json_geometry_name = b'"geometry_name":%b,' % orjson.dumps(name)
        else:
            json_geometry_name = b""

        return b'    {"type":"Feature","id":%b,%b"geometry":%b,"properties":%b}' % (
            orjson.dumps(f"{feature_type.name}.{instance.pk}"),
            json_geometry_name,
            self.render_geometry(projection, instance),
            orjson.dumps(properties, default=_json_default),
        )

    def _format_geojson_value(self, value):
        if isinstance(value, datetime):
            return value.astimezone(timezone.utc)
        elif isinstance(value, models.Model):
            # ForeignKey, not defined as complex type.
            return str(value)
        else:
            return value

    def render_geometry(self, projection: FeatureProjection, instance: models.Model) -> bytes:
        """Generate the proper GeoJSON notation for a geometry.
        This calls the GDAL C-API rendering found in 'GEOSGeometry.json'
        """
        geometry = projection.get_main_geometry_value(instance)
        if geometry is None:
            return b"null"

        # The GeoJSON spec requires coordinates to be x,y (longitude,latitude),
        # so web-based clients don't have to ship projection tables.
        projection.output_crs.apply_to(geometry, axis_order=AxisOrder.TRADITIONAL)
        return geometry.json.encode()

    def get_header(self) -> dict:
        """Generate the header fields.

        The format is based on the WFS 3.0 DRAFT. The count fields are moved
        to the footer allowing them to be calculated without performing queries.
        """
        output_crs = self.collection.results[0].projection.output_crs
        return {
            "type": "FeatureCollection",
            "timeStamp": self._format_geojson_value(self.collection.timestamp),
            # "numberReturned": is written at the end for better query performance.
            "crs": {"type": "name", "properties": {"name": str(output_crs)}},
        }

    def get_footer(self) -> dict:
        """Generate the last fields of the response.
        By moving the links, numberReturned/numberMatched fields to the end,
        it's not always needed to perform queries to calculate these.

        The format is based on the WFS 3.0 DRAFT which defines the extra
        pagination headers and numberMatched/numberReturned headers.
        While the draft suggests to put these fields first, there is no such
        requirement in JSON.
        """
        return {
            "links": self.get_links(),
            "numberReturned": self.collection.number_returned,
            "numberMatched": self.collection.number_matched,
        }

    def get_links(self) -> list:
        """Generate the pagination links"""
        links = []
        if self.collection.next:
            links.append(
                {
                    "href": self.collection.next,
                    "rel": "next",
                    "type": "application/geo+json",
                    "title": "next page",
                }
            )
        if self.collection.previous:
            links.append(
                {
                    "href": self.collection.previous,
                    "rel": "previous",
                    "type": "application/geo+json",
                    "title": "previous page",
                }
            )
        return links

    def get_properties(
        self, projection: FeatureProjection, xsd_elements: list[XsdElement], instance: models.Model
    ) -> dict:
        """Collect the data for the 'properties' field.

        This is based on the original XSD definition,
        so the rendering is consistent with other output formats.

        :param projection: Overview of all fields to render.
        :param xsd_elements: The XSD elements to render.
        :param instance: The object instance.
        """
        props = {}
        for xsd_element in xsd_elements:
            if not xsd_element.type.is_geometry:
                value = xsd_element.get_value(instance)
                if xsd_element.type.is_complex_type:
                    # Nested object data
                    if value is None:
                        props[xsd_element.name] = None
                    else:
                        sub_elements = projection.xsd_child_nodes[xsd_element]
                        if xsd_element.is_many:
                            # "..._to_many relation; reverse FK, M2M or array field.
                            props[xsd_element.name] = [
                                self.get_properties(projection, sub_elements, item)
                                for item in value
                            ]
                        else:
                            props[xsd_element.name] = self.get_properties(
                                projection, sub_elements, value
                            )
                else:
                    # Scalar value, or list (for ArrayField).
                    props[xsd_element.name] = self._format_geojson_value(value)

        return props


class DBGeoJsonRenderer(GeoJsonRenderer):
    """GeoJSON renderer that relays the geometry rendering to the database.

    This is even more efficient than calling the C-API for each feature.
    """

    def decorate_queryset(self, projection: FeatureProjection, queryset):
        """Update the queryset to let the database render the GML output.
        This is far more efficient than GeoDjango's logic, which performs a
        C-API call for every single coordinate of a geometry.
        """
        queryset = super().decorate_queryset(projection, queryset)
        # If desired, the entire FeatureCollection could be rendered
        # in PostgreSQL as well: https://postgis.net/docs/ST_AsGeoJSON.html
        main_geo_element = projection.feature_type.main_geometry_element
        if main_geo_element is not None:
            queryset = queryset.defer(main_geo_element.orm_path).annotate(
                _as_db_geojson=AsGeoJSON(
                    get_db_geometry_target(main_geo_element, projection.output_crs),
                    precision=conf.GISSERVER_DB_PRECISION,
                )
            )

        return queryset

    def render_geometry(self, projection: FeatureProjection, instance: models.Model) -> bytes:
        """Generate the proper GeoJSON notation for a geometry"""
        # Database server rendering
        if projection.main_geometry_element is None:
            return b"null"

        geojson = instance._as_db_geojson
        return b"null" if geojson is None else geojson.encode()



from __future__ import annotations

import logging
import math
import typing
from collections.abc import Iterator
from io import BytesIO, StringIO
from itertools import chain

from django.conf import settings
from django.db import models
from django.http import HttpResponse, StreamingHttpResponse
from django.http.response import HttpResponseBase  # Django 3.2 import location

from gisserver.exceptions import wrap_filter_errors
from gisserver.features import FeatureType
from gisserver.parsers.values import fix_type_name
from gisserver.parsers.xml import split_ns
from gisserver.types import XsdAnyType, XsdNode

from .utils import render_xmlns_attributes, to_qname

logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from gisserver.operations.base import WFSOperation
    from gisserver.projection import FeatureProjection, FeatureRelation

    from .results import FeatureCollection, SimpleFeatureCollection


class OutputRenderer:
    """Base class for rendering content.

    Note most rendering logic will generally
    need to use :class:`~gisserver.output.XmlOutputRenderer`
    or :class:`~gisserver.output.CollectionOutputRenderer` as their base class
    """

    #: Default content type for the HTTP response
    content_type = "application/octet-stream"

    def __init__(self, operation: WFSOperation):
        """Base method for all output rendering."""
        self.operation = operation

    def get_response(self) -> HttpResponseBase:
        """Render the output as regular or streaming response."""
        stream = self.render_stream()
        if isinstance(stream, (str, bytes, StringIO, BytesIO)):
            # Not a real stream, output anyway as regular HTTP response.
            return HttpResponse(
                content=stream,
                content_type=self.content_type,
                headers=self.get_headers(),
            )
        else:
            # An actual generator.
            # Peek the generator so initial exceptions can still be handled,
            # and get rendered as normal HTTP responses with the proper status.
            try:
                start = next(stream)  # peek, so any raised OWSException here is handled by OWSView
                stream = chain([start], self._trap_exceptions(stream))
            except StopIteration:
                pass

            # Handover to WSGI server (starts streaming when reading the contents)
            return StreamingHttpResponse(
                streaming_content=stream,
                content_type=self.content_type,
                headers=self.get_headers(),
            )

    def get_headers(self) -> dict[str, str]:
        """Override to define HTTP headers to add."""
        return {}

    def _trap_exceptions(self, stream):
        """Decorate the generator to show exceptions"""
        try:
            yield from stream
        except Exception as e:
            # Can't return 500 at this point,
            # but can still tell the client what happened.
            yield self.render_exception(e)
            raise

    def render_exception(self, exception: Exception):
        """Inform the client that the stream processing was interrupted with an exception.
        The exception can be rendered in the format fits with the output.

        Purposefully, not much information is given, so avoid informing clients.
        The actual exception is still raised and logged server-side.
        """
        if settings.DEBUG:
            return f"{exception.__class__.__name__}: {exception}"
        else:
            return f"{exception.__class__.__name__} during rendering!"

    def render_stream(self):
        """Implement this in subclasses to implement a custom output format.

        The implementation may return a ``str``/``bytes`` object, which becomes
        a normal ``HttpResponse`` object **OR** return a generator
        that emits chunks. Such generator is wrapped in a ``StreamingHttpResponse``.
        """
        raise NotImplementedError()


class XmlOutputRenderer(OutputRenderer):
    """Base class/mixin for XML-based rendering.

    This provides the logic to translate XML elements into QName aliases.
    """

    #: Default content type for the HTTP response
    content_type = "text/xml; charset=utf-8"

    #: Default extra namespaces to include in the xmlns="..." attributes, and use for to_qname().
    xml_namespaces = {}

    def __init__(self, operation: WFSOperation):
        """Base method for all output rendering."""
        super().__init__(operation)
        self.app_namespaces = {
            **self.xml_namespaces,
            **operation.view.get_xml_namespaces_to_prefixes(),
        }

    def render_xmlns_attributes(self):
        """Render XML Namespace declaration attributes"""
        return render_xmlns_attributes(self.app_namespaces)

    def to_qname(self, xsd_type: XsdNode | XsdAnyType, namespaces=None) -> str:
        """Generate the aliased name for the element or type."""
        return to_qname(xsd_type.namespace, xsd_type.name, namespaces or self.app_namespaces)

    def feature_to_qname(self, feature_type: FeatureType | str) -> str:
        """Convert the FeatureType name to a QName"""
        if isinstance(feature_type, FeatureType):
            return to_qname(feature_type.xml_namespace, feature_type.name, self.app_namespaces)
        else:
            # e.g. "return_type" for StoredQueryDescription/QueryExpressionText
            type_name = fix_type_name(feature_type, self.operation.view.xml_namespace)
            ns, localname = split_ns(type_name)
            return to_qname(ns, localname, self.app_namespaces)


class CollectionOutputRenderer(OutputRenderer):
    """Base class to create streaming responses."""

    #: Allow to override the maximum page size.
    #: This value can be 'math.inf' to support endless pages by default.
    max_page_size = None

    #: An optional content-disposition header to output
    content_disposition = None

    def __init__(self, operation: WFSOperation, collection: FeatureCollection):
        """
        Receive the collected data to render.

        :param operation: The calling WFS Operation (e.g. GetFeature class)
        :param collection: The collected data for rendering.
        """
        super().__init__(operation)
        self.collection = collection
        self.apply_projection()

    def apply_projection(self):
        """Perform presentation-layer logic enhancements on all results.
        This calls :meth:`decorate_queryset` for
        each :class:`~gisserver.output.SimpleFeatureCollection`.
        """
        for sub_collection in self.collection.results:
            queryset = self.decorate_queryset(
                projection=sub_collection.projection,
                queryset=sub_collection.queryset,
            )
            if sub_collection._result_cache is not None:
                raise RuntimeError(
                    "SimpleFeatureCollection QuerySet was already processed in output rendering."
                )
            if queryset is not None:
                sub_collection.queryset = queryset

    def decorate_queryset(
        self, projection: FeatureProjection, queryset: models.QuerySet
    ) -> models.QuerySet:
        """Apply presentation layer logic to the queryset.

        This allows fine-tuning the queryset for any special needs of the output rendering type.

        :param projection: The projection information, including feature that is being rendered.
        :param queryset: The constructed queryset so far.
        """
        if queryset._result_cache is not None:
            raise RuntimeError(
                "QuerySet was already processed before passing to the output rendering."
            )

        # Avoid fetching relations, fetch these within the same query,
        prefetches = self._get_prefetch_related(projection)
        if prefetches:
            logger.debug(
                "QuerySet for %s prefetches: %r",
                queryset.model._meta.label,
                [p.prefetch_through for p in prefetches],
            )
            queryset = queryset.prefetch_related(*prefetches)

        logger.debug(
            "QuerySet for %s only retrieves: %r",
            queryset.model._meta.label,
            projection.only_fields,
        )
        return queryset.only("pk", *projection.only_fields)

    def _get_prefetch_related(self, projection: FeatureProjection) -> list[models.Prefetch]:
        """Summarize which fields read data from relations.

        This combines the input from flattened and complex fields,
        in the unlikely case both variations are used in the same feature.
        """
        # When PROPERTYNAME is used, determine all ORM paths (and levels) that this query will touch.
        # The prefetch will only be applied when its field coincide with this list.

        return [
            models.Prefetch(
                orm_relation.orm_path,
                queryset=self.get_prefetch_queryset(projection, orm_relation),
            )
            for orm_relation in projection.orm_relations
        ]

    def get_prefetch_queryset(
        self, projection: FeatureProjection, feature_relation: FeatureRelation
    ) -> models.QuerySet | None:
        """Generate a custom queryset that's used to prefetch a relation."""
        # Multiple elements could be referencing the same model, just take first that is filled in.
        if feature_relation.related_model is None:
            return None

        # This will also apply .only() based on the projection's feature_relation
        return projection.feature_type.get_related_queryset(feature_relation)

    def get_headers(self):
        """Return the response headers"""
        if self.content_disposition:
            # Offer a common quick content-disposition logic that works for all possible queries.
            return {
                "Content-Disposition": self.content_disposition.format(
                    **self.get_content_disposition_kwargs()
                )
            }

        return {}

    def get_content_disposition_kwargs(self) -> dict:
        """Offer a common quick content-disposition logic that works for all possible queries."""
        sub_collection = self.collection.results[0]
        if sub_collection.stop == math.inf:
            page = f"{sub_collection.start}-end" if sub_collection.start else "all"
        elif sub_collection.stop:
            page = f"{sub_collection.start}-{sub_collection.stop - 1}"
        else:
            page = "results"

        return {
            "typenames": "+".join(
                feature_type.name
                for sub in self.collection.results
                for feature_type in sub.feature_types
            ),
            "page": page,
            "date": self.collection.date.strftime("%Y-%m-%d %H.%M.%S%z"),
            "timestamp": self.collection.timestamp,
        }

    def read_features(self, sub_collection: SimpleFeatureCollection) -> Iterator[models.Model]:
        """A wrapper to read features from a collection, while raising WFS exceptions on query errors."""
        with wrap_filter_errors(sub_collection.source_query):
            yield from sub_collection



import math

from django.conf import settings
from django.core.signals import setting_changed
from django.dispatch import receiver

_originals = {}

# -- feature flags

# Configure whether the <ows:WGS84BoundingBox> should be included in GetCapabilities.
# This is an expensive operation to calculate, hence it reduces the overall performance.
GISSERVER_CAPABILITIES_BOUNDING_BOX = getattr(
    settings, "GISSERVER_CAPABILITIES_BOUNDING_BOX", True
)

# Whether to use the database for rendering GML / GeoJSON fragments.
# This gives a better performance overall, but output may vary between database vendors.
GISSERVER_USE_DB_RENDERING = getattr(settings, "GISSERVER_USE_DB_RENDERING", True)

# The precision to use for DB rendering. (PostGIS stores reliably up till 15 decimals)
GISSERVER_DB_PRECISION = getattr(settings, "GISSERVER_DB_PRECISION", 15)

# Whether to strictly check whether the provided CRS is accepted.
# Otherwise, all database-supported SRID's are allowed.
GISSERVER_SUPPORTED_CRS_ONLY = getattr(settings, "GISSERVER_SUPPORTED_CRS_ONLY", True)

# Whether the total results need to be counted.
# By disabling this, clients just need to fetch more pages
# 0 = No counting, 1 = all pages, 2 = only for the first page.
GISSERVER_COUNT_NUMBER_MATCHED = getattr(settings, "GISSERVER_COUNT_NUMBER_MATCHED", 1)

# -- output rendering

# Following https://docs.geoserver.org/stable/en/user/services/wfs/axis_order.html here:
# Whether the older EPSG:4326 notation (instead of the OGC recommended styles)
# should render in legacy longitude/latitude (x/y) ordering.
# This increases interoperability with legacy web clients,
# as others use urn:ogc:def:crs:EPSG::4326 or http://www.opengis.net/def/crs/epsg/0/4326.
GISSERVER_FORCE_XY_EPSG_4326 = getattr(settings, "GISSERVER_FORCE_XY_EPSG_4326", True)

# Whether the legacy CRS notation http://www.opengis.net/gml/srs/epsg.xml# should render in X/Y
GISSERVER_FORCE_XY_OLD_CRS = getattr(settings, "GISSERVER_FORCE_XY_OLD_CRS", True)

# Extra output formats for GetFeature (see documentation for details)
GISSERVER_EXTRA_OUTPUT_FORMATS = getattr(settings, "GISSERVER_EXTRA_OUTPUT_FORMATS", {})
GISSERVER_GET_FEATURE_OUTPUT_FORMATS = getattr(
    settings, "GISSERVER_GET_FEATURE_OUTPUT_FORMATS", {}
)

# -- max page size

# Allow tuning the page size without having to override code.
# This corresponds with the "DefaultMaxFeatures" setting.
GISSERVER_DEFAULT_MAX_PAGE_SIZE = getattr(settings, "GISSERVER_DEFAULT_MAX_PAGE_SIZE", 5000)

# CSV exports have a higher default page size, as these results can be streamed.
GISSERVER_GEOJSON_MAX_PAGE_SIZE = getattr(settings, "GISSERVER_GEOJSON_MAX_PAGE_SIZE", math.inf)
GISSERVER_CSV_MAX_PAGE_SIZE = getattr(settings, "GISSERVER_CSV_MAX_PAGE_SIZE", math.inf)

# -- debugging

# Whether to follow the WFS standards strictly (breaks CITE conformance testing)
GISSERVER_WFS_STRICT_STANDARD = getattr(settings, "GISSERVER_WFS_STRICT_STANDARD", False)

# Whether to wrap filter errors in a nice response, or raise an exception
GISSERVER_WRAP_FILTER_DB_ERRORS = getattr(settings, "GISSERVER_WRAP_FILTER_DB_ERRORS", True)


@receiver(setting_changed)
def _on_settings_change(setting, value, enter, **kwargs):
    if not setting.startswith("GISSERVER_"):
        return

    conf_module = globals()
    if value is None and not enter:
        # override_settings().disable() returns what the django settings module had.
        # Revert to our defaults here instead.
        value = _originals.get(setting)
    else:
        # Track defaults of this file for reverting to them
        _originals.setdefault(setting, conf_module[setting])

    conf_module[setting] = value



"""The main configuration for exposing model data in the WFS server.

The "feature type" definitions define what models and attributes are exposed in the WFS server.
When a model attribute is mentioned in the feature type, it can be exposed and queried against.
Any field that is not mentioned in a definition, will therefore not be available, nor queryable.
This metadata is used in the ``GetCapabilities`` call to advertise all available feature types.

The "feature type" definitions ares translated internally into
an internal XML Schema Definition (made from :mod:`gisserver.types`).
That schema maps all model attributes to a specific XML layout, and includes
all XSD Complex Types, elements and attributes linked to the Django model metadata.

The feature type classes (and field types) offer a flexible translation
from attribute listings into a schema definition.
For example, model relationships can be modelled to a different XML layout.
"""

from __future__ import annotations

import html
import itertools
import logging
from dataclasses import dataclass
from functools import cached_property, lru_cache
from typing import TYPE_CHECKING, Literal

from django.contrib.gis.db import models as gis_models
from django.contrib.gis.db.models import GeometryField
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db import models
from django.http import HttpRequest

from gisserver import conf
from gisserver.compat import ArrayField, GeneratedField
from gisserver.crs import CRS
from gisserver.db import get_wgs84_bounding_box
from gisserver.exceptions import ExternalValueError, InvalidParameterValue
from gisserver.geometries import WGS84BoundingBox
from gisserver.parsers.xml import parse_qname, xmlns
from gisserver.types import (
    GeometryXsdElement,
    GmlBoundedByElement,
    GmlIdAttribute,
    GmlNameElement,
    XPathMatch,
    XsdAnyType,
    XsdComplexType,
    XsdElement,
    XsdTypes,
)

if TYPE_CHECKING:
    from gisserver.projection import FeatureRelation

__all__ = [
    "FeatureType",
    "field",
    "FeatureField",
    "ComplexFeatureField",
]

logger = logging.getLogger(__name__)

XSD_TYPES = {
    models.CharField: XsdTypes.string,
    models.TextField: XsdTypes.string,
    models.BooleanField: XsdTypes.boolean,
    models.IntegerField: XsdTypes.integer,
    models.PositiveIntegerField: XsdTypes.nonNegativeInteger,
    models.PositiveBigIntegerField: XsdTypes.nonNegativeInteger,
    models.PositiveSmallIntegerField: XsdTypes.nonNegativeInteger,
    models.AutoField: XsdTypes.integer,  # Only as of Django 3.0 this extends from IntegerField
    models.FloatField: XsdTypes.double,
    models.DecimalField: XsdTypes.decimal,
    models.TimeField: XsdTypes.time,
    models.DateTimeField: XsdTypes.dateTime,  # note: DateTimeField extends DateField!
    models.DateField: XsdTypes.date,
    models.DurationField: XsdTypes.duration,
    models.URLField: XsdTypes.anyURI,
    gis_models.PointField: XsdTypes.gmlPointPropertyType,
    gis_models.PolygonField: XsdTypes.gmlSurfacePropertyType,
    gis_models.LineStringField: XsdTypes.gmlCurvePropertyType,
    gis_models.MultiPointField: XsdTypes.gmlMultiPointPropertyType,
    gis_models.MultiPolygonField: XsdTypes.gmlMultiSurfacePropertyType,
    gis_models.MultiLineStringField: XsdTypes.gmlMultiCurvePropertyType,
    # Generic alternatives
    gis_models.GeometryCollectionField: XsdTypes.gmlMultiGeometryPropertyType,
    gis_models.GeometryField: XsdTypes.gmlGeometryPropertyType,
}
DEFAULT_XSD_TYPE = XsdTypes.anyType


def _get_basic_field_type(
    field_name: str, model_field: models.Field | models.ForeignObjectRel
) -> XsdAnyType:
    """Determine the XSD field type for a Django field."""
    if ArrayField is not None and isinstance(model_field, ArrayField):
        # Determine the type based on the contents.
        # The array notation is written as "is_many"
        model_field = model_field.base_field

    if GeneratedField is not None and isinstance(model_field, GeneratedField):
        # Allow things like: models.GeneratedField(SomeFunction("geofield"), output_field=models.GeometryField())
        model_field = model_field.output_field

    try:
        # Direct instance, quickly resolved!
        return XSD_TYPES[model_field.__class__]
    except KeyError:
        pass

    if isinstance(model_field, models.ForeignKey):
        # Don't let it query on the relation value yet
        return _get_basic_field_type(field_name, model_field.target_field)
    elif isinstance(model_field, models.ForeignObjectRel):
        # e.g. ManyToOneRel descriptor of a foreignkey_id field.
        return _get_basic_field_type(field_name, model_field.remote_field.target_field)
    else:
        # Subclass checks:
        for field_cls, xsd_type in XSD_TYPES.items():
            if isinstance(model_field, field_cls):
                return xsd_type

    # Default:
    if isinstance(model_field, GeometryField):
        return XsdTypes.gmlGeometryPropertyType
    else:
        # Default XML choice:
        return DEFAULT_XSD_TYPE


def _get_model_fields(
    model: type[models.Model],
    fields: list[str] | Literal["__all__"],
    parent: ComplexFeatureField | None = None,
    feature_type: FeatureType | None = None,
):
    if fields == "__all__":
        # All regular fields
        # Relationships will not be expanded since it can expose so many other fields.
        # Only the relationships that store an ID on the model itself will be exposed.
        return [
            FeatureField(
                name=f.attname,  # using attname so foreignkeys use the name_id field.
                # .bind() is called directly by providing these arguments via init:
                model=model,
                parent=parent,
                feature_type=feature_type,
            )
            for f in model._meta.get_fields()
            if not f.is_relation or f.many_to_one or f.one_to_one  # ForeignKey, OneToOneField
        ]
    else:
        # Only defined fields
        fields = [f if isinstance(f, FeatureField) else FeatureField(f) for f in fields]
        for field in fields:  # type: FeatureField
            field.bind(model, parent=parent, feature_type=feature_type)
        return fields


@dataclass
class ServiceDescription:
    """Basic metadata for an exposed GIS service."""

    title: str
    abstract: str | None = None
    keywords: list[str] | None = None

    provider_name: str | None = None
    provider_site: str | None = None
    contact_person: str | None = None


class FeatureField:
    """The configuration for a field inside a WFS Feature.

    This defines how a Django model field is mapped into
    an XSD definition that the remaining application uses.
    """

    #: Allow to override the XSD element type that this field will generate.
    xsd_element_class: type[XsdElement] = None

    model: type[models.Model] | None
    model_field: models.Field | models.ForeignObjectRel | None

    def __init__(
        self,
        name,
        model_attribute=None,
        model=None,
        parent: ComplexFeatureField | None = None,
        feature_type: FeatureType | None = None,
        abstract=None,
        xsd_class: type[XsdElement] | None = None,
    ):
        """
        Initialize a single field of the feature.

        :param name: Name of the model field.
        :param model_attribute: Which model attribute to access. This can be a dotted field path.
        :param model: Which model is accessed. Usually this is passed via :meth:`bind`.
        :param parent: The parent field of this element. Usually this is passed via :meth:`bind`.
        :param feature_type: The feature this field is a part of. Usually this is passed via :meth:`bind`.
        :param abstract: The "help text" or short abstract/description for this field.
        :param xsd_class: Override which class is used to construct the internal XsdElement.
                          This controls the schema rendering, value retrieval and rendering of the element.
        """
        self.name = name
        self.model_attribute = model_attribute
        self.model = None
        self.model_field = None
        self.parent = parent
        self.feature_type = feature_type
        self.abstract = abstract

        # Allow to override the class attribute on 'self',
        # which avoids having to subclass this field class as well.
        if xsd_class is not None:
            self.xsd_element_class = xsd_class

        self._nillable_relation = False
        self._nillable_output_field = False
        if model is not None:
            self.bind(model, parent=parent, feature_type=feature_type)

    def __repr__(self):
        if self.model_field is None:
            return f"<{self.__class__.__name__}: {self.name} (unbound)>"
        else:
            return (
                f"<{self.__class__.__name__}: {self.name}, source={self.absolute_model_attribute}>"
            )

    def _get_xsd_type(self):
        return _get_basic_field_type(self.name, self.model_field)

    def bind(
        self,
        model: type[models.Model],
        parent: ComplexFeatureField | None = None,
        feature_type: FeatureType | None = None,
    ):
        """Late-binding for the model.

        This resolves the model field object from the provided model.
        This method is called internally when the field definition wasn't
        linked to a model yet. This allows the fields to be defined first,
        in external code, and then become part of the ``FeatureType`` fields
        list.

        :param model: The model is field is linked to.
        :param parent: When this element is part of a complex feature,
                       this links to the parent field.
        :param feature_type: The original feature type that his element was mentioned in.
        """
        if self.model is not None:
            raise RuntimeError(f"Feature field '{self.name}' cannot be reused")
        self.model = model
        self.parent = parent
        self.feature_type = feature_type

        if self.model_attribute:
            # Support dot based field traversal.
            field_path = self.model_attribute.split(".")
            try:
                field: models.Field = self.model._meta.get_field(field_path[0])

                for name in field_path[1:]:
                    if field.null:
                        self._nillable_relation = True

                    if not field.is_relation:
                        # Note this tests the previous loop variable, so it checks whether the
                        # field can be walked into in order to resolve the next dotted name below.
                        raise ImproperlyConfigured(
                            f"FeatureField '{self.name}' has an invalid model_attribute: "
                            f"field '{name}' is a '{field.__class__.__name__}', not a relation."
                        )

                    field = field.related_model._meta.get_field(name)
                self.model_field = field
            except FieldDoesNotExist as e:
                raise ImproperlyConfigured(
                    f"FeatureField '{self.name}' has an invalid"
                    f" model_attribute: '{self.model_attribute}' can't be"
                    f" resolved for model '{self.model.__name__}'."
                ) from e
        else:
            try:
                self.model_field = self.model._meta.get_field(self.name)
                if GeneratedField is not None and isinstance(self.model_field, GeneratedField):
                    self._nillable_output_field = self.model_field.output_field.null
            except FieldDoesNotExist as e:
                raise ImproperlyConfigured(
                    f"FeatureField '{self.name}' can't be resolved for model '{self.model.__name__}'."
                    " Either set 'model_attribute' or change the name."
                ) from e

    @cached_property
    def absolute_model_attribute(self) -> str:
        """Determine the full attribute of the field."""
        if self.model_field is None:
            raise RuntimeError(f"bind() was not called for {self!r}")

        if self.parent is not None:
            return f"{self.parent.absolute_model_attribute}.{self.model_attribute or self.name}"
        else:
            return self.model_attribute or self.name

    @cached_property
    def xsd_element(self) -> XsdElement:
        """Define the XMLSchema definition for a model field.

        This definition is used by the remaining application to access the
        data. It's the basis for DescribeFeatureType, and it's ``get_value()``
        method is read to access the model field data.
        """
        if self.model_field is None:
            raise RuntimeError(f"bind() was not called for {self!r}")

        xsd_type = self._get_xsd_type()

        # Determine max number of occurrences.
        if xsd_type.is_geometry:
            max_occurs = 1  # be explicit here, like mapserver does.
        elif self.model_field.many_to_many or self.model_field.one_to_many:
            max_occurs = "unbounded"  # M2M or reverse FK field
        elif ArrayField is not None and isinstance(self.model_field, ArrayField):
            max_occurs = self.model_field.size or "unbounded"
        else:
            max_occurs = None  # default is 1, but attribute can be left out.

        # Determine which subclass to use for the element.
        xsd_element_class = self.xsd_element_class or (
            GeometryXsdElement if xsd_type.is_geometry else XsdElement
        )

        return xsd_element_class(
            name=self.name,
            type=xsd_type,
            namespace=self.feature_type.xml_namespace,  # keep all types in the same namespace for now.
            nillable=(
                self.model_field.null or self._nillable_relation or self._nillable_output_field
            ),
            min_occurs=0,
            max_occurs=max_occurs,
            model_attribute=self.model_attribute,
            absolute_model_attribute=self.absolute_model_attribute,
            source=self.model_field,
            feature_type=self.feature_type,
        )


class ComplexFeatureField(FeatureField):
    """The configuration for an embedded relation field.

    This field type is suitable for any relational object, including
    foreign keys, reverse relations and M2M fields. The internal logic
    translates the relation into an embedded XSD complex type.
    """

    def __init__(
        self,
        name: str,
        fields: list[str | FeatureField] | Literal["__all__"],
        model_attribute=None,
        model=None,
        abstract=None,
        xsd_class=None,
        xsd_base_type=None,
    ):
        """
        :param name: Name of the model field.
        :param fields: If the field exposes a foreign key, provide its child element names.
            This can be a list of :func:`field` elements, or plain field names.
            Using ``__all__`` also works but is not recommended outside testing.
        :param model_attribute: Which model attribute to access. This can be a dotted field path.
        :param abstract: The "help text" or short abstract/description for this field.
        :param xsd_class: Override which class is used to construct the internal XsdElement.
                          This controls the schema rendering, value retrieval and rendering of the element.
        :param xsd_base_type: Override which class is the base class for the element.
        """
        super().__init__(
            name,
            model_attribute=model_attribute,
            model=model,
            abstract=abstract,
            xsd_class=xsd_class,
        )
        self.xsd_base_type = xsd_base_type
        self._fields = fields

    @cached_property
    def fields(self) -> list[FeatureField]:
        """Provide all fields that will be rendered as part of this complex field."""
        return _get_model_fields(
            self.target_model, self._fields, parent=self, feature_type=self.feature_type
        )

    def _get_xsd_type(self) -> XsdComplexType:
        """Generate the XSD description for the field with an object relation."""
        pk_field = self.target_model._meta.pk

        return XsdComplexType(
            name=f"{self.target_model._meta.object_name}Type",
            namespace=self.feature_type.xml_namespace,  # keep all types in the same namespace for now.
            elements=[field.xsd_element for field in self.fields],
            attributes=[
                # Add gml:id attribute definition, so it can be resolved in xpath
                GmlIdAttribute(
                    type_name=self.name,
                    source=pk_field,
                    model_attribute=pk_field.name,
                    feature_type=self.feature_type,
                )
            ],
            base=self.xsd_base_type,
            source=self.target_model,
        )

    @property
    def target_model(self) -> type[models.Model]:
        """Detect which model the relation points to."""
        if self.model_field is None:
            raise RuntimeError("FeatureField.bind() is not called yet")
        elif not self.model_field.is_relation:
            raise ImproperlyConfigured(
                f"{self.__class__.__name__} does not support fields of "
                f"type {self.model_field.__class__.__name__}."
            )
        else:
            return self.model_field.related_model


def field(
    name: str,
    *,
    model_attribute=None,
    abstract: str | None = None,
    fields: list[str | FeatureField] | Literal["__all__"] | None = None,
    xsd_class: type[XsdElement] | None = None,
) -> FeatureField:
    """Shortcut to define a WFS field.

    This automatically selects the proper field class,
    so little knowledge is needed about the internal working.

    :param name: Name of the model field.
    :param model_attribute: Which model attribute to access. This can be a dotted field path.
    :param abstract: The "help text" or short abstract/description for this field.
    :param fields: If the field exposes a foreign key, provide its child element names.
        This can be a list of :func:`field` elements, or plain field names.
        Using ``__all__`` also works but is not recommended outside testing.
    :param xsd_class: Override which class is used to construct the internal XsdElement.
                      This controls the schema rendering, value retrieval and rendering of the element.
    """
    if fields is not None:
        return ComplexFeatureField(
            name=name,
            model_attribute=model_attribute,
            fields=fields,
            abstract=abstract,
            xsd_class=xsd_class,
        )
    else:
        return FeatureField(
            name,
            model_attribute=model_attribute,
            abstract=abstract,
            xsd_class=xsd_class,
        )


class FeatureType:
    """Declare a feature that is exposed on the map.

    All WFS operations use this class to read the feature type.
    You may subclass this class to provide extensions,
    such as redefining :meth:`get_queryset`.

    This corresponds with a single Django model.
    """

    #: Allow to override the XSD complex type that this feature will generate.
    xsd_type_class: type[XsdComplexType] = XsdComplexType

    def __init__(
        self,
        queryset: models.QuerySet,
        *,
        fields: list[str | FeatureField] | Literal["__all__"] | None = None,
        display_field_name: str | None = None,
        geometry_field_name: str | None = None,
        name: str | None = None,
        # WFS Metadata:
        title: str | None = None,
        abstract: str | None = None,
        keywords: list[str] | None = None,
        crs: CRS | None = None,
        other_crs: list[CRS] | None = None,
        metadata_url: str | None = None,
        # Settings
        show_name_field: bool = True,
        xml_namespace: str | None = None,
    ):
        """
        :param queryset: The queryset to retrieve the data.
        :param fields: Define which fields to show in the WFS data.
            This can be a list of field names, or :class:`FeatureField` objects.
        :param display_field_name: Name of the field that's used as general string representation.
        :param geometry_field_name: Name of the geometry field to expose (default = auto-detect).
        :param name: Name, also used as XML tag name.
        :param title: Used in WFS metadata.
        :param abstract: Used in WFS metadata.
        :param keywords: Used in WFS metadata.
        :param crs: Used in WFS metadata.
        :param other_crs: Used in WFS metadata.
        :param metadata_url: Used in WFS metadata.
        :param show_name_field: Whether to show the ``gml:name`` or the GeoJSON ``geometry_name``
            field. Default is to show a field when ``name_field`` is given.
        :param xml_namespace: The XML namespace to use, will be set by :meth:`bind_namespace` otherwise.
        """
        if isinstance(queryset, models.QuerySet):
            self.queryset = queryset
            self.model = queryset.model
        elif isinstance(queryset, type) and issubclass(queryset, models.Model):
            # In case a model is provided, fix that
            self.model = queryset
            self.queryset = self.model.objects.all()
        else:
            raise TypeError("FeatureType expects a model or queryset")

        self._fields = fields
        self.display_field_name = display_field_name
        self._geometry_field_name = geometry_field_name
        self.name = name or self.model._meta.model_name
        self.title = title or self.model._meta.verbose_name
        self.abstract = abstract
        self.keywords = keywords or []
        self._crs = crs
        self.other_crs = other_crs or []
        self.metadata_url = metadata_url

        # Settings
        self.show_name_field = show_name_field
        self.xml_namespace = xml_namespace

        # Validate that the name doesn't require XML escaping.
        if html.escape(self.name) != self.name or " " in self.name or ":" in self.name:
            raise ValueError(f"Invalid feature name for XML: <{self.xml_name}>")

        self._cached_resolver = lru_cache(200)(self._inner_resolve_element)

    def __repr__(self):
        return (
            f"<{self.__class__.__qualname__}: {self.xml_name},"
            f" fields={self.fields!r},"
            f" geometry_field_name={self.main_geometry_element.absolute_model_attribute!r}>"
        )

    def bind_namespace(self, default_xml_namespace: str):
        """Make sure the feature type receives the settings from the parent view."""
        if not self.xml_namespace:
            self.xml_namespace = default_xml_namespace

    def check_permissions(self, request: HttpRequest):
        """Hook that allows subclasses to reject access for datasets.
        It may raise a Django PermissionDenied error.

        This can check for example whether ``request.user`` may access this feature.

        The parsed WFS request is available as ``request.ows_request``.
        Currently, this can be a :class:`~gisserver.parsers.wfs20.GetFeature`
        or :class:`~gisserver.parsers.wfs20.GetPropertyValue` instance.
        """

    @cached_property
    def xml_name(self) -> str:
        """Return the feature tag as XML Full Qualified name"""
        return f"{{{self.xml_namespace}}}{self.name}" if self.xml_namespace else self.name

    @cached_property
    def supported_crs(self) -> list[CRS]:
        """Return all spatial reference system ID's that this feature supports."""
        return [self.crs] + self.other_crs

    @cached_property
    def all_geometry_elements(self) -> list[GeometryXsdElement]:
        """Provide access to all geometry elements from *all* nested levels."""
        return self.xsd_type.geometry_elements + list(
            itertools.chain.from_iterable(
                # Take the geometry elements at each object level.
                xsd_element.type.geometry_elements
                for xsd_element in self.xsd_type.all_complex_elements
                # The 'None' level is the root node, which is already added before.
                # For now, don't support geometry fields on an M2M relation.
                # If that use-case is needed, it would require additional work to implement.
                if xsd_element is not None and not xsd_element.is_many
            )
        )

    @cached_property
    def _local_model_field_names(self) -> list[str]:
        """Tell which local fields of the model will be accessed by this feature."""
        return [
            (ff.model_field.name if ff.model_field else ff.model_attribute)
            for ff in self.fields
            if not ff.model_field.many_to_many
            and not ff.model_field.one_to_many
            and not (ff.model_attribute and "." in ff.model_attribute)
        ]

    @cached_property
    def fields(self) -> list[FeatureField]:
        """Define which fields to render."""
        if (
            self._geometry_field_name
            and "." in self._geometry_field_name
            and (self._fields is None or self._fields == "__all__")
        ):
            # If you want to define more complex relationships, please be explicit in which fields you want.
            # Allowing autoconfiguration of this is complex and likely not what you're looking for either.
            raise ImproperlyConfigured(
                "Using a geometry_field_name path requires defining 'fields' explicitly."
            )

        # This lazy reading allows providing 'fields' as lazy value.
        if self._fields is None:
            # Autoconfig, no fields defined.
            if not self._geometry_field_name:
                self._geometry_field_name = next(
                    f.name
                    for f in self.model._meta.get_fields()
                    if isinstance(f, gis_models.GeometryField)
                )

            return [FeatureField(self._geometry_field_name, model=self.model, feature_type=self)]
        else:
            # Either __all__ or an explicit list.
            return _get_model_fields(self.model, self._fields, feature_type=self)

    @cached_property
    def display_field(self) -> models.Field | None:
        """Give access to the field that holds the string-representation."""
        if not self.show_name_field or not self.display_field_name:
            return None
        else:
            return self.model._meta.get_field(self.display_field_name)

    @cached_property
    def main_geometry_element(self) -> GeometryXsdElement:
        """Give access to the main geometry element."""
        if not self._geometry_field_name:
            try:
                # Take the first element that has a geometry.
                return self.all_geometry_elements[0]
            except IndexError:
                raise ImproperlyConfigured(
                    f"FeatureType '{self.name}' does not expose any geometry field "
                    f"as part of its definition."
                ) from None
        else:
            try:
                return next(
                    e
                    for e in self.all_geometry_elements
                    if e.absolute_model_attribute == self._geometry_field_name
                )
            except StopIteration:
                raise ImproperlyConfigured(
                    f"FeatureType '{self.name}' does not expose the geometry field "
                    f"'{self._geometry_field_name}' as part of its definition."
                ) from None

    @cached_property
    def crs(self) -> CRS:
        """Tell which projection the data should be presented at."""
        if self._crs is None:
            # Default CRS
            return CRS.from_srid(self.main_geometry_element.source_srid)  # checks lookup too
        else:
            return self._crs

    def get_queryset(self) -> models.QuerySet:
        """Return the queryset that is used as basis for this feature."""
        # Return a queryset that only retrieves the fields that are actually displayed.
        # That that without .only(), use at least `self.queryset.all()` so a clone is returned.
        logger.debug(
            "QuerySet for %s default only retrieves: %r",
            self.queryset.model._meta.label,
            self._local_model_field_names,
        )
        return self.queryset.only(*self._local_model_field_names)

    def get_related_queryset(self, feature_relation: FeatureRelation) -> models.QuerySet:
        """Return the queryset that is used for prefetching related data."""
        if feature_relation.related_model is None:
            raise RuntimeError(
                f"Unable to create prefetch queryset for relation {feature_relation.orm_path}, "
                f"source model is not defined for: {feature_relation.xsd_elements!r}"
            )
        # Return a queryset that only retrieves the fields that are displayed.
        logger.debug(
            "QuerySet for %s by default only retrieves: %r",
            feature_relation.related_model._meta.label,
            feature_relation._local_model_field_names,
        )
        queryset = feature_relation.related_model.objects.only(
            *feature_relation._local_model_field_names
        )
        return self.filter_related_queryset(queryset)  # Allow overriding by FeatureType subclasses

    def filter_related_queryset(self, queryset: models.QuerySet) -> models.QuerySet:
        """When a related object returns a queryset, this hook allows extra filtering."""
        return queryset

    def get_bounding_box(self) -> WGS84BoundingBox | None:
        """Returns a WGS84 BoundingBox for the complete feature.

        This is used by the GetCapabilities request. It may return ``None``
        when the database table is empty, or the custom queryset doesn't
        return any results.

        Note that the ``<ows:WGS84BoundingBox>`` element always uses longitude/latitude,
        as it doesn't describe a CRS.
        """
        if not self.main_geometry_element:
            return None

        return get_wgs84_bounding_box(self.get_queryset(), self.main_geometry_element)

    def get_display_value(self, instance: models.Model) -> str:
        """Generate the display name value"""
        if self.display_field_name:
            return getattr(instance, self.display_field_name)
        else:
            return str(instance)

    @property
    def xsd_base_type(self) -> XsdAnyType:
        """Return the base class for the :attr:`xsd_type`.

        This builds an XsdComplexType element that represents
        the contents of :attr:`XsdTypes.gmlAbstractFeatureType`.
        By providing this as Complex Type, the filters can also resolve the
        attributes of ``@gml:id``, ``<gml:name>`` and ``<gml:boundedBy>`` nodes.
        """
        pk_field = self.model._meta.pk

        # Define <gml:boundedBy>
        base_elements = [GmlBoundedByElement(feature_type=self)]
        if self.show_name_field:
            # Add <gml:name>
            gml_name = GmlNameElement(
                model_attribute=self.display_field_name,
                source=self.display_field,
                feature_type=self,
            )
            base_elements.insert(0, gml_name)

        # Write out the definition of XsdTypes.gmlAbstractFeatureType as actual class definition.
        # By having these base elements, the XPath queries can also resolve these like any other feature field elements.
        return XsdComplexType(
            name="AbstractFeatureType",
            namespace=xmlns.gml.value,
            elements=base_elements,
            attributes=[
                # Add gml:id="..." attribute definition so it can be resolved in xpath
                GmlIdAttribute(
                    type_name=self.name,
                    source=pk_field,
                    model_attribute=pk_field.name,
                    feature_type=self,
                )
            ],
            base=XsdTypes.gmlAbstractGMLType,
        )

    @cached_property
    def xsd_type(self) -> XsdComplexType:
        """Return the definition of this feature as an XSD Complex Type."""
        return self.xsd_type_class(
            name=f"{self.name[0].upper()}{self.name[1:]}Type",
            elements=[field.xsd_element for field in self.fields],
            namespace=self.xml_namespace,
            base=self.xsd_base_type,
            source=self.model,
        )

    def resolve_element(self, xpath: str, ns_aliases: dict[str, str]) -> XPathMatch:
        """Resolve the element, and the matching object.

        This is used to convert XPath references in requests
        to the actual elements and model attributes for queries.

        Internally, this method caches results.
        """
        # When the XML POST request used xmlns="http://www.opengis.net/wfs/2.0",
        # this will become the default namespace elements are translated into.
        # However, the XPath elements should be interpreted within our application namespace instead.
        # When the default namespace is missing, it should also resolve to our feature.
        ns_aliases = ns_aliases.copy()
        ns_aliases[""] = self.xml_namespace

        if len(ns_aliases) > 10:
            # Avoid filling memory by caching large namespace blobs.
            nodes = self._inner_resolve_element(xpath, ns_aliases)
        else:
            # Go through lru_cache() for faster lookup of the same elements.
            # Note 1: the cache will be less effective when clients use different namespace aliases.
            # Note 2: when WFSView overrides get_feature_types() the cache may only exist per request.
            nodes = self._cached_resolver(xpath, HDict(ns_aliases))

        if nodes is None:
            raise ExternalValueError(f"Field '{xpath}' does not exist.")

        return XPathMatch(self, nodes, query=xpath)

    def _inner_resolve_element(self, xpath: str, ns_aliases: dict[str, str]):
        """Inner part of resolve_element() that is cached.
        This performs any additional checks that happen at the root-level only.
        """
        # Avoid complex XPath features for now. Only element/child works.
        # The attribute selector is not fully implemented, but tested elsewhere.
        if "//" in xpath:
            raise NotImplementedError(
                f"XPath selectors with deeper descendant selectors are not supported: {xpath}"
            )
        elif "::" in xpath:
            raise NotImplementedError(
                f"XPath selectors with expanded syntax are not supported: {xpath}"
            )
        elif "(" in xpath:
            raise NotImplementedError(f"XPath selectors with functions are not supported: {xpath}")

        # Allow /app:ElementName/.. as "absolute" path, simply strip it for now.
        # This is only actually needed to support join queries, which we don't.
        xpath = xpath.lstrip("/")
        root_name, _, _ = xpath.partition("/")
        root_xml_name = parse_qname(root_name, ns_aliases)
        if root_xml_name == self.xml_name:
            xpath = xpath[len(root_name) + 1 :]

        return self.xsd_type.resolve_element_path(xpath, ns_aliases)

    def resolve_crs(self, crs: CRS, locator="") -> CRS:
        """Check a parsed CRS against the list of supported types."""
        for candidate in self.supported_crs:
            # Not using self.supported_crs.index(crs), as that depends on CRS.__eq__():
            if candidate.matches(crs, compare_legacy=False):
                if candidate.force_xy != crs.force_xy:
                    # user provided legacy CRS, allow output in legacy CRS
                    return crs
                else:
                    # Replace the parsed CRS with the declared one.
                    return candidate

        # No match found
        if conf.GISSERVER_SUPPORTED_CRS_ONLY:
            raise InvalidParameterValue(
                f"Feature '{self.name}' does not support CRS '{crs}'.",
                locator=locator,
            ) from None
        else:
            return crs


class HDict(dict):
    """Dict that can be used in lru_cache()."""

    def __hash__(self):
        return hash(frozenset(self.items()))



"""Output rendering logic.

Note that the Django format_html() / mark_safe() logic is not used here,
as it's quite a performance improvement to just use html.escape().

We've tried replacing this code with lxml and that turned out to be much slower.
As some functions will be called 5000x, this code is also designed to avoid making
much extra method calls per field. Some bits are non-DRY inlined for this reason.
"""

from __future__ import annotations

import itertools
import re
from collections import defaultdict
from datetime import date, datetime, time, timezone
from decimal import Decimal as D
from io import StringIO
from operator import itemgetter
from typing import cast

from django.contrib.gis import geos
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.http import HttpResponse

from gisserver.crs import CRS84
from gisserver.db import (
    AsGML,
    get_db_geometry_target,
    get_db_rendered_geometry,
    get_geometries_union,
    replace_queryset_geometries,
)
from gisserver.exceptions import NotFound, WFSException
from gisserver.geometries import BoundingBox
from gisserver.parsers.xml import xmlns
from gisserver.projection import FeatureProjection, FeatureRelation
from gisserver.types import GeometryXsdElement, GmlBoundedByElement, XsdElement, XsdNode, XsdTypes

from .base import CollectionOutputRenderer, XmlOutputRenderer
from .results import SimpleFeatureCollection
from .utils import (
    attr_escape,
    tag_escape,
    value_to_text,
    value_to_xml_string,
)

GML_RENDER_FUNCTIONS = {}
RE_SRS_NAME = re.compile(r'srsName="([^"]+)"')


def register_geos_type(geos_type):
    def _inc(func):
        GML_RENDER_FUNCTIONS[geos_type] = func
        return func

    return _inc


class GML32Renderer(CollectionOutputRenderer, XmlOutputRenderer):
    """Render the GetFeature XML output in GML 3.2 format"""

    content_type = "text/xml; charset=utf-8"
    content_disposition = 'inline; filename="{typenames} {page} {date}.xml"'
    xml_collection_tag = "FeatureCollection"
    xml_sub_collection_tag = "FeatureCollection"  # Mapserver does not use SimpleFeatureCollection
    chunk_size = 40_000
    gml_seq = 0

    # Aliases to use for XML namespaces
    xml_namespaces = {
        "http://www.opengis.net/wfs/2.0": "wfs",
        "http://www.opengis.net/gml/3.2": "gml",
        "http://www.w3.org/2001/XMLSchema-instance": "xsi",  # for xsi:nil="true" and xsi:schemaLocation.
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.feature_types = [
            feature_type
            for sub_collection in self.collection.results
            for feature_type in sub_collection.feature_types
        ]

        self.build_namespace_map()

    def build_namespace_map(
        self,
    ):
        """Collect all namespaces which namespaces are used by features."""
        try:
            # Make aliases for the feature type elements.
            self.feature_qnames = {
                feature_type: self.feature_to_qname(feature_type)
                for feature_type in self.feature_types
            }
            self.xml_qnames = self._build_xml_qnames()
        except KeyError as e:
            raise ImproperlyConfigured(f"No XML namespace alias defined in WFSView for {e}") from e

    def _build_xml_qnames(self) -> dict[XsdNode, str]:
        """Collect aliases for all rendered elements.
        This uses a defaultdict so optional attributes (e.g. by GetPropertyValue) can also be resolved.
        The dict lookup also speeds up, no need to call ``node_to_qname()`` for each tag.
        """
        known_tags = {
            xsd_element: self.to_qname(xsd_element)
            for sub_collection in self.collection.results
            for xsd_element in sub_collection.projection.all_elements
        }
        return defaultdict(self.to_qname, known_tags)

    def get_response(self):
        """Render the output as streaming response."""
        if self.collection.results and self.collection.results[0].projection.output_standalone:
            # WFS spec requires that GetFeatureById output only returns the contents.
            # The streaming response is avoided here, to allow returning a 404.
            return self.get_by_id_response()
        else:
            # Use default streaming response, with render_stream()
            return super().get_response()

    def get_by_id_response(self):
        """Render a standalone item, for GetFeatureById"""
        sub_collection = self.collection.results[0]
        sub_collection.source_query.finalize_results(sub_collection)  # Allow 404
        self.start_collection(sub_collection)

        instance = sub_collection.first()
        if instance is None:
            raise NotFound("Feature not found.")

        self.app_namespaces.pop(xmlns.wfs20.value)  # not rendering wfs tags.
        self.output = StringIO()
        self._write = self.output.write
        self.write_by_id_response(
            sub_collection, instance, extra_xmlns=f" {self.render_xmlns_attributes()}"
        )
        content = self.output.getvalue()
        return HttpResponse(content, content_type=self.content_type)

    def write_by_id_response(self, sub_collection: SimpleFeatureCollection, instance, extra_xmlns):
        """Default behavior for standalone response is writing a feature (can be changed by GetPropertyValue)"""
        self._write('<?xml version="1.0" encoding="UTF-8"?>\n')
        self.write_feature(
            projection=sub_collection.projection,
            instance=instance,
            extra_xmlns=extra_xmlns,
        )

    def render_xsi_schema_location(self):
        """Render the value for the xsi:schemaLocation="..." block."""
        # Find which namespaces are exposed
        types_by_namespace = defaultdict(list)
        for feature_type in self.feature_types:
            types_by_namespace[feature_type.xml_namespace].append(feature_type)

        # Schema location are pairs of "{namespace-uri} {schema-url} {namespace-uri2} {schema-url2}"
        # WFS server types refer to the endpoint that generates the XML Schema for the given feature.
        schema_locations = []
        for xml_namespace, feature_types in types_by_namespace.items():
            schema_locations.append(xml_namespace)
            schema_locations.append(self.operation.view.get_xml_schema_url(feature_types))

        schema_locations += [
            "http://www.opengis.net/wfs/2.0 http://schemas.opengis.net/wfs/2.0/wfs.xsd",
            "http://www.opengis.net/gml/3.2 http://schemas.opengis.net/gml/3.2.1/gml.xsd",
        ]
        return " ".join(schema_locations)

    def render_exception(self, exception: Exception):
        """Render the exception in a format that fits with the output.

        The WSF XSD spec has a hidden gem: an ``<wfs:truncatedResponse>`` element
        can be rendered at the end of a feature collection
        to inform the client an error happened during rendering.
        """
        message = super().render_exception(exception)
        buffer = self.output.getvalue()

        # Wrap into <ows:ExceptionReport> tag.
        if not isinstance(exception, WFSException):
            exception = WFSException(message, code=exception.__class__.__name__, status_code=500)
        exception.debug_hint = False

        # Only at the top-level, an exception report can be rendered, close any remaining tags.
        if len(self.collection.results) > 1:
            sub_closing = f"</wfs:{self.xml_sub_collection_tag}>\n</wfs:member>\n"
            if not buffer.endswith(sub_closing):
                buffer += sub_closing

        return (
            f"{buffer}"
            "  <wfs:truncatedResponse>"
            f"{exception.as_xml()}"
            "  </wfs:truncatedResponse>\n"
            f"</wfs:{self.xml_collection_tag}>\n"
        )

    def render_stream(self):
        """Render the XML as streaming content.
        This renders the standard <wfs:FeatureCollection> / <wfs:ValueCollection>
        """
        collection = self.collection
        self.output = output = StringIO()
        self._write = self.output.write

        # The base class peaks the generator and handles early exceptions.
        # Any database exceptions during calculating the number of results
        # are all handled by the main WFS view.
        number_matched = collection.number_matched
        number_matched = int(number_matched) if number_matched is not None else "unknown"
        number_returned = collection.number_returned

        next = previous = ""
        if collection.next:
            next = f' next="{attr_escape(collection.next)}"'
        if collection.previous:
            previous = f' previous="{attr_escape(collection.previous)}"'

        self._write(
            f"""<?xml version='1.0' encoding="UTF-8" ?>\n"""
            f"<wfs:{self.xml_collection_tag} {self.render_xmlns_attributes()}"
            f' xsi:schemaLocation="{attr_escape(self.render_xsi_schema_location())}"'
            f' timeStamp="{collection.timestamp}"'
            f' numberMatched="{number_matched}"'
            f' numberReturned="{int(number_returned)}"'
            f"{next}{previous}>\n"
        )

        if number_returned:
            has_multiple_collections = len(collection.results) > 1

            for sub_collection in collection.results:
                projection = sub_collection.projection
                if projection.output_crs.force_xy and projection.output_crs.is_north_east_order:
                    self._write(
                        "<!--\n"
                        f" NOTE: you are requesting the legacy projection notation '{tag_escape(projection.output_crs.origin)}'."
                        f" Please use '{tag_escape(projection.output_crs.urn)}' instead.\n\n"
                        " This also means output coordinates are ordered in legacy the 'west, north' axis ordering.\n"
                        "\n-->\n"
                    )

                self.start_collection(sub_collection)
                if has_multiple_collections:
                    self._write(
                        f"<wfs:member>\n"
                        f"<wfs:{self.xml_sub_collection_tag}"
                        f' timeStamp="{collection.timestamp}"'
                        f' numberMatched="{int(sub_collection.number_matched)}"'
                        f' numberReturned="{int(sub_collection.number_returned)}">\n'
                    )

                for instance in self.read_features(sub_collection):
                    self.gml_seq = 0  # need to increment this between write_xml_field calls
                    self._write("<wfs:member>\n")
                    self.write_feature(projection, instance)
                    self._write("</wfs:member>\n")

                    # Only perform a 'yield' every once in a while,
                    # as it goes back-and-forth for writing it to the client.
                    if output.tell() > self.chunk_size:
                        xml_chunk = output.getvalue()
                        output.seek(0)
                        output.truncate(0)
                        yield xml_chunk

                if has_multiple_collections:
                    self._write(f"</wfs:{self.xml_sub_collection_tag}>\n</wfs:member>\n")

        self._write(f"</wfs:{self.xml_collection_tag}>\n")
        yield output.getvalue()

    def start_collection(self, sub_collection: SimpleFeatureCollection):
        """Hook to allow initialization per feature type"""

    def write_feature(
        self, projection: FeatureProjection, instance: models.Model, extra_xmlns=""
    ) -> None:
        """Write the contents of the object value.

        This output is typically wrapped in <wfs:member> tags
        unless it's used for a GetPropertyById response.
        """
        feature_type = projection.feature_type
        feature_xml_qname = self.feature_qnames[feature_type]

        # Write <app:FeatureTypeName> start node
        pk = tag_escape(str(instance.pk))
        self._write(f'<{feature_xml_qname} gml:id="{feature_type.name}.{pk}"{extra_xmlns}>\n')
        # Write all fields, both base class and local elements.
        for xsd_element in projection.xsd_root_elements:
            # Note that writing 5000 features with 30 tags means this code make 150.000 method calls.
            # Hence, branching to different rendering styles is branched here instead of making such
            # call into a generic "write_field()" function and stepping out from there.

            if xsd_element.type.is_geometry:
                # Separate call which can be optimized (no need to overload write_xml_field() for all calls).
                self.write_gml_field(projection, xsd_element, instance)
            else:
                value = xsd_element.get_value(instance)
                if xsd_element.is_many:
                    self.write_many(projection, xsd_element, value)
                else:
                    # e.g. <gml:name>, or all other <app:...> nodes.
                    self.write_xml_field(projection, xsd_element, value)

        self._write(f"</{feature_xml_qname}>\n")

    def write_many(self, projection: FeatureProjection, xsd_element: XsdElement, value) -> None:
        """Write a node that has multiple values (e.g. array or queryset)."""
        # some <app:...> node that has multiple values
        if value is None:
            # No tag for optional element (see PropertyIsNull), otherwise xsi:nil node.
            if xsd_element.min_occurs:
                xml_qname = self.xml_qnames[xsd_element]
                self._write(f'<{xml_qname} xsi:nil="true"/>\n')
        else:
            for item in value:
                self.write_xml_field(projection, xsd_element, value=item)

    def write_xml_field(
        self, projection: FeatureProjection, xsd_element: XsdElement, value, extra_xmlns=""
    ):
        """Write the value of a single field."""
        xml_qname = self.xml_qnames[xsd_element]
        if value is None:
            self._write(f'<{xml_qname} xsi:nil="true"{extra_xmlns}/>\n')
        elif xsd_element.type.is_complex_type:
            # Expanded foreign relation / dictionary
            self.write_xml_complex_type(projection, xsd_element, value, extra_xmlns=extra_xmlns)
        else:
            # As this is likely called 150.000 times during a request, this is optimized.
            # Avoided a separate call to value_to_xml_string() and avoided isinstance() here.
            value_cls = value.__class__
            if value_cls is str:  # most cases
                value = tag_escape(value)
            elif value_cls is datetime:
                value = value.astimezone(timezone.utc).isoformat()
            elif value_cls is bool:
                value = "true" if value else "false"
            elif (
                value_cls is not int
                and value_cls is not float
                and value_cls is not D
                and value_cls is not date
                and value_cls is not time
            ):
                # Non-string or custom field that extended a scalar.
                # Any of the other types have a faster f"{value}" translation that produces the correct text.
                value = value_to_xml_string(value)

            self._write(f"<{xml_qname}{extra_xmlns}>{value}</{xml_qname}>\n")

    def write_xml_complex_type(
        self, projection: FeatureProjection, xsd_element: XsdElement, value, extra_xmlns=""
    ) -> None:
        """Write a single field, that consists of sub elements"""
        xml_qname = self.xml_qnames[xsd_element]
        self._write(f"<{xml_qname}{extra_xmlns}>\n")
        for sub_element in projection.xsd_child_nodes[xsd_element]:
            if sub_element.type.is_geometry:
                # Separate call which can be optimized (no need to overload write_xml_field() for all calls).
                self.write_gml_field(projection, sub_element, value)
            else:
                sub_value = sub_element.get_value(value)
                if sub_element.is_many:
                    self.write_many(projection, sub_element, sub_value)
                else:
                    self.write_xml_field(projection, sub_element, sub_value)
        self._write(f"</{xml_qname}>\n")

    def write_gml_field(
        self,
        projection: FeatureProjection,
        geo_element: GeometryXsdElement,
        instance: models.Model,
        extra_xmlns="",
    ) -> None:
        """Separate method to allow overriding this for db-performance optimizations."""
        if geo_element.type is XsdTypes.gmlBoundingShapeType:
            # Special case for <gml:boundedBy>, which doesn't need xsd:nil values, nor a gml:id.
            # The value is not a GEOSGeometry either, as that is not exposed by django.contrib.gis.
            envelope = cast(GmlBoundedByElement, geo_element).get_value(
                instance, crs=projection.output_crs
            )
            if envelope is not None:
                self._write(self.render_gml_bounds(envelope))
        else:
            # Regular geometry elements.
            xml_qname = self.xml_qnames[geo_element]
            value = geo_element.get_value(instance)
            if value is None:
                # Avoid incrementing gml_seq
                self._write(f'<{xml_qname} xsi:nil="true"{extra_xmlns}/>\n')
            else:
                gml_id = self.get_gml_id(instance._meta.object_name, instance.pk)

                # the following is somewhat faster, but will render GML 2, not GML 3.2:
                # gml = value.ogr.gml
                # pos = gml.find(">")  # Will inject the gml:id="..." tag.
                # gml = f"{gml[:pos]} gml:id="{attr_escape(gml_id)}"{gml[pos:]}"

                gml = self.render_gml_value(projection, gml_id, value)
                self._write(f"<{xml_qname}{extra_xmlns}>{gml}</{xml_qname}>\n")

    def get_gml_id(self, prefix: str, object_id) -> str:
        """Generate the gml:id value, which is required for GML 3.2 objects."""
        self.gml_seq += 1
        return f"{prefix}.{object_id}.{self.gml_seq}"

    def render_gml_bounds(self, envelope: BoundingBox) -> str:
        """Render the gml:boundedBy element that contains an Envelope.
        This uses an internal object type,
        as :mod:`django.contrib.gis.geos` only provides a 4-tuple envelope.
        """
        return f"""<gml:boundedBy><gml:Envelope srsDimension="2" srsName="{attr_escape(str(envelope.crs))}">
            <gml:lowerCorner>{envelope.min_x} {envelope.min_y}</gml:lowerCorner>
            <gml:upperCorner>{envelope.max_x} {envelope.max_y}</gml:upperCorner>
          </gml:Envelope></gml:boundedBy>\n"""

    def render_gml_value(
        self,
        projection: FeatureProjection,
        gml_id,
        value: geos.GEOSGeometry | None,
        extra_xmlns="",
    ) -> str:
        """Normal case: 'value' is raw geometry data.."""
        # In case this is a standalone response, this will be the top-level element, hence includes the xmlns.
        base_attrs = f' gml:id="{attr_escape(gml_id)}" srsName="{attr_escape(str(projection.output_crs))}"{extra_xmlns}'
        projection.output_crs.apply_to(value)
        return self._render_gml_type(value, base_attrs=base_attrs)

    def _render_gml_type(self, value: geos.GEOSGeometry, base_attrs=""):
        """Render an GML value (this is also called from MultiPolygon)."""
        try:
            # Avoid isinstance checks, do a direct lookup
            method = GML_RENDER_FUNCTIONS[value.__class__]
        except KeyError:
            return f"<!-- No rendering implemented for {value.geom_type} -->"
        return method(self, value, base_attrs=base_attrs)

    @register_geos_type(geos.Point)
    def render_gml_point(self, value: geos.Point, base_attrs=""):
        coords = " ".join(map(str, value.coords))
        dim = 3 if value.hasz else 2
        return (
            f"<gml:Point{base_attrs}>"
            f'<gml:pos srsDimension="{dim}">{coords}</gml:pos>'
            f"</gml:Point>"
        )

    @register_geos_type(geos.Polygon)
    def render_gml_polygon(self, value: geos.Polygon, base_attrs=""):
        # lol: http://erouault.blogspot.com/2014/04/gml-madness.html
        ext_ring = self.render_gml_linear_ring(value.exterior_ring)
        buf = StringIO()
        buf.write(f"<gml:Polygon{base_attrs}><gml:exterior>{ext_ring}</gml:exterior>")
        for i in range(value.num_interior_rings):
            buf.write("<gml:interior>")
            buf.write(self.render_gml_linear_ring(value[i + 1]))
            buf.write("</gml:interior>")
        buf.write("</gml:Polygon>")
        return buf.getvalue()

    @register_geos_type(geos.GeometryCollection)
    def render_gml_multi_geometry(self, value: geos.GeometryCollection, base_attrs):
        children = "".join(self._render_gml_type(child) for child in value)
        return f"<gml:MultiGeometry{base_attrs}>{children}</gml:MultiGeometry>"

    @register_geos_type(geos.MultiPolygon)
    def render_gml_multi_polygon(self, value: geos.GeometryCollection, base_attrs):
        children = "".join(self._render_gml_type(child) for child in value)
        return f"<gml:MultiPolygon{base_attrs}>{children}</gml:MultiPolygon>"

    @register_geos_type(geos.MultiLineString)
    def render_gml_multi_line_string(self, value: geos.MultiPoint, base_attrs):
        children = "</gml:lineStringMember><gml:lineStringMember>".join(
            self.render_gml_line_string(child) for child in value
        )
        return (
            f"<gml:MultiLineString{base_attrs}>"
            f"<gml:lineStringMember>{children}</gml:lineStringMember>"
            f"</gml:MultiLineString>"
        )

    @register_geos_type(geos.MultiPoint)
    def render_gml_multi_point(self, value: geos.MultiPoint, base_attrs):
        children = "</gml:pointMember><gml:pointMember>".join(
            self.render_gml_point(child) for child in value
        )
        return (
            f"<gml:MultiPoint{base_attrs}>"
            f"<gml:pointMember>{children}</gml:pointMember>"
            f"</gml:MultiPoint>"
        )

    @register_geos_type(geos.LinearRing)
    def render_gml_linear_ring(self, value: geos.LinearRing, base_attrs=""):
        # NOTE: this is super slow. value.tuple performs a C-API call for every point!
        coords = " ".join(map(str, itertools.chain.from_iterable(value.tuple)))
        dim = "3" if value.hasz else "2"
        # <gml:coordinates> is still valid in GML3, but deprecated (part of GML2).
        return (
            f"<gml:LinearRing{base_attrs}>"
            f'<gml:posList srsDimension="{dim}">{coords}</gml:posList>'
            "</gml:LinearRing>"
        )

    @register_geos_type(geos.LineString)
    def render_gml_line_string(self, value: geos.LineString, base_attrs=""):
        # NOTE: this is super slow. value.tuple performs a C-API call for every point!
        coords = " ".join(map(str, itertools.chain.from_iterable(value.tuple)))
        dim = "3" if value.hasz else "2"
        return (
            f"<gml:LineString{base_attrs}>"
            f'<gml:posList srsDimension="{dim}">{coords}</gml:posList>'
            "</gml:LineString>"
        )


class DBGMLRenderingMixin:

    def render_gml_value(self, projection, gml_id, value: str, extra_xmlns=""):
        """DB optimized: 'value' is pre-rendered GML XML string."""
        # Write the gml:id inside the first tag
        end_pos = value.find(">")
        gml_tag = value[:end_pos]
        id_pos = gml_tag.find("gml:id=")
        if id_pos == -1:
            # Inject
            gml = f'{gml_tag} gml:id="{attr_escape(gml_id)}"{extra_xmlns}{value[end_pos:]}'
        else:
            # Replace
            end_pos1 = gml_tag.find('"', id_pos + 8)
            gml = (
                f"{gml_tag[:id_pos]}"
                f'gml:id="{attr_escape(gml_id)}'
                f"{value[end_pos1:end_pos]}"  # from " right until >
                f"{extra_xmlns}"  # extra namespaces?
                f"{value[end_pos:]}"  # from > and beyond
            )

        return self._fix_db_crs_name(projection, gml)

    def _fix_db_crs_name(self, projection: FeatureProjection, gml: str) -> str:
        if projection.output_crs.force_xy:
            # When legacy output is used, make sure the srsName matches the input.
            # PostgreSQL will still generate the EPSG:xxxx notation.
            return gml.replace(
                'srsName="EPSG:', 'srsName="http://www.opengis.net/gml/srs/epsg.xml#', 1
            )
        elif projection.output_crs == CRS84:
            # Fix PostgreSQL not knowing it's CRS84 or WGS84 (both srid 4326)
            return gml.replace(
                'srsName="urn:ogc:def:crs:EPSG::4326"', 'srsName="urn:ogc:def:crs:OGC::CRS84"', 1
            )
        else:
            return gml


class DBGML32Renderer(DBGMLRenderingMixin, GML32Renderer):
    """Faster GetFeature renderer that uses the database to render GML 3.2"""

    def decorate_queryset(self, projection: FeatureProjection, queryset: models.QuerySet):
        """Update the queryset to let the database render the GML output.
        This is far more efficient than GeoDjango's logic, which performs a
        C-API call for every single coordinate of a geometry.
        """
        queryset = super().decorate_queryset(projection, queryset)

        # Retrieve gml:boundedBy in pre-rendered format.
        if projection.has_bounded_by:
            queryset = queryset.annotate(
                _as_envelope_gml=self.get_db_envelope_as_gml(projection, queryset),
            )

        # Retrieve geometries as pre-rendered instead.
        # Only take the geometries of the current level.
        # The annotations for relations will be handled by prefetches and get_prefetch_queryset()
        use_modern = not projection.output_crs.force_xy
        return replace_queryset_geometries(
            queryset,
            projection.geometry_elements,
            projection.output_crs,
            AsGML,
            is_latlon=use_modern and projection.output_crs.is_north_east_order,
            long_urn=use_modern,
        )

    def get_prefetch_queryset(
        self,
        projection: FeatureProjection,
        feature_relation: FeatureRelation,
    ) -> models.QuerySet | None:
        """Perform DB annotations for prefetched relations too."""
        queryset = super().get_prefetch_queryset(projection, feature_relation)
        if queryset is None:
            return None

        # Find which fields are GML elements
        use_modern = not projection.output_crs.force_xy
        return replace_queryset_geometries(
            queryset,
            feature_relation.geometry_elements,
            projection.output_crs,
            AsGML,
            is_latlon=use_modern and projection.output_crs.is_north_east_order,
            long_urn=use_modern,
        )

    def get_db_envelope_as_gml(self, projection: FeatureProjection, queryset) -> AsGML:
        """Offload the GML rendering of the envelope to the database.

        This also avoids offloads the geometry union calculation to the DB.
        """
        geo_fields_union = self._get_geometries_union(projection, queryset)
        use_modern = not projection.output_crs.force_xy
        return AsGML(
            geo_fields_union,
            envelope=True,
            is_latlon=use_modern and projection.output_crs.is_north_east_order,
            long_urn=use_modern,
        )

    def _get_geometries_union(self, projection: FeatureProjection, queryset):
        """Combine all geometries of the model in a single SQL function."""
        # Apply transforms where needed, in case some geometries use a different SRID.
        return get_geometries_union(
            [
                get_db_geometry_target(geo_element, output_crs=projection.output_crs)
                for geo_element in projection.all_geometry_elements
                if geo_element.source is not None  # excludes GmlBoundedByElement
            ],
            using=queryset.db,
        )

    def write_gml_field(
        self,
        projection: FeatureProjection,
        geo_element: GeometryXsdElement,
        instance: models.Model,
        extra_xmlns="",
    ) -> None:
        """Write the value of an GML tag.

        This optimized version takes a pre-rendered XML from the database query.
        """
        xml_qname = self.xml_qnames[geo_element]
        if geo_element.type is XsdTypes.gmlBoundingShapeType:
            gml = instance._as_envelope_gml
            if gml is None:
                return

            gml = self._fix_db_crs_name(projection, gml)
        else:
            value = get_db_rendered_geometry(instance, geo_element, AsGML)
            if value is None:
                # Avoid incrementing gml_seq, make nil tag.
                self._write(f'<{xml_qname} xsi:nil="true"{extra_xmlns}/>\n')
                return

            # Get gml tag to write as value.
            gml_id = self.get_gml_id(instance._meta.object_name, instance.pk)
            gml = self.render_gml_value(projection, gml_id, value, extra_xmlns=extra_xmlns)

        self._write(f"<{xml_qname}{extra_xmlns}>{gml}</{xml_qname}>\n")


class GML32ValueRenderer(GML32Renderer):
    """Render the GetPropertyValue XML output in GML 3.2 format.

    Geoserver seems to generate the element tag inside each ``<wfs:member>`` element. We've applied this one.
    The GML standard demonstrates to render only their content inside a ``<wfs:member>`` element
    (either plain text or an ``<gml:...>`` tag). Not sure what is right here.
    """

    content_type = "text/xml; charset=utf-8"
    content_type_plain = "text/plain; charset=utf-8"
    content_disposition = 'inline; filename="{typenames} {page} {date}-value.xml"'
    content_disposition_plain = 'inline; filename="{typenames} {page} {date}-value.txt"'
    xml_collection_tag = "ValueCollection"
    xml_sub_collection_tag = "ValueCollection"
    _escape_value = staticmethod(value_to_xml_string)
    gml_value_getter = itemgetter("member")

    def decorate_queryset(self, projection: FeatureProjection, queryset: models.QuerySet):
        # Don't optimize queryset, it only retrieves one value
        # The data is already limited to a ``queryset.values()`` in ``QueryExpression.get_queryset()``.
        return queryset

    def write_by_id_response(
        self, sub_collection: SimpleFeatureCollection, instance: dict, extra_xmlns
    ):
        """The value rendering only renders the value. not a complete feature"""
        if sub_collection.projection.property_value_node.is_attribute:
            # Output as plain text
            self.content_type = self.content_type_plain  # change for this instance!
            self.content_disposition = self.content_disposition_plain
            self._escape_value = value_to_text  # avoid XML escaping
        else:
            self._write('<?xml version="1.0" encoding="UTF-8"?>\n')

        # Write the single tag, no <wfs:member> around it.
        self.write_feature(sub_collection.projection, instance, extra_xmlns=extra_xmlns)

    def write_feature(self, projection: FeatureProjection, instance: dict, extra_xmlns="") -> None:
        """Write the XML for a single object.
        In this case, it's only a single XML tag.
        """
        xsd_node = projection.property_value_node
        if xsd_node.type.is_geometry:
            self.write_gml_field(
                projection,
                cast(GeometryXsdElement, xsd_node),
                instance,
                extra_xmlns=extra_xmlns,
            )
        elif xsd_node.is_attribute:
            value = instance["member"]
            if value is not None:
                value = xsd_node.format_raw_value(instance["member"])  # for gml:id
                value = self._escape_value(value)
                self._write(value)
        elif xsd_node.is_array:
            if (value := instance["member"]) is not None:
                # <wfs:member> doesn't allow multiple items as children, for new render as separate members.
                xml_qname = self.xml_qnames[xsd_node]
                first = True
                for item in value:
                    if item is not None:
                        if not first:
                            self._write("</wfs:member>\n<wfs:member>")

                        item = self._escape_value(item)
                        self._write(f"<{xml_qname}>{item}</{xml_qname}>\n")
                        first = False
        elif xsd_node.type.is_complex_type:
            raise NotImplementedError("GetPropertyValue with complex types is not implemented")
            # self.write_xml_complex_type(projection, self.xsd_node, instance['member'])
        else:
            self.write_xml_field(
                projection,
                cast(XsdElement, xsd_node),
                value=instance["member"],
                extra_xmlns=extra_xmlns,
            )

    def write_gml_field(
        self,
        projection: FeatureProjection,
        geo_element: GeometryXsdElement,
        instance: dict,
        extra_xmlns="",
    ) -> None:
        """Overwritten to allow dict access instead of model access."""
        if geo_element.type is XsdTypes.gmlBoundingShapeType:
            raise NotImplementedError(
                "rendering <gml:boundedBy> in GetPropertyValue is not implemented."
            )

        value = self.gml_value_getter(instance)  # "member" or "gml_member"
        xml_qname = self.xml_qnames[geo_element]
        if value is None:
            # Avoid incrementing gml_seq
            self._write(f'<{xml_qname} xsi:nil="true"{extra_xmlns}/>\n')
            return

        gml_id = self.get_gml_id(geo_element.source.model._meta.object_name, instance["pk"])
        gml = self.render_gml_value(projection, gml_id, value, extra_xmlns=extra_xmlns)
        self._write(f"<{xml_qname}{extra_xmlns}>{gml}</{xml_qname}>\n")


class DBGML32ValueRenderer(DBGMLRenderingMixin, GML32ValueRenderer):
    """Faster GetPropertyValue renderer that uses the database to render GML 3.2"""

    gml_value_getter = itemgetter("gml_member")

    def decorate_queryset(self, projection: FeatureProjection, queryset):
        """Update the queryset to let the database render the GML output."""
        # As this is a classmethod, self.value_reference is not available yet.
        element = projection.property_value_node
        if element.type.is_geometry:
            # Add 'gml_member' to point to the pre-rendered GML version.
            geo_element = cast(GeometryXsdElement, element)
            use_modern = not projection.output_crs.force_xy
            return queryset.values(
                "pk",
                gml_member=AsGML(
                    get_db_geometry_target(geo_element, projection.output_crs),
                    is_latlon=use_modern and projection.output_crs.is_north_east_order,
                    long_urn=use_modern,
                ),
            )
        else:
            return queryset



"""All parser logic to process incoming XML data.

This handles all tags, including:

* ``<wfs:...>``
* ``<fes:...>``
* ``<gml:...>``

Internally, the XML string is translated into an Abstract Syntax Tree (AST).
These objects contain to logic to process each bit of the XML tree.
The GET request parameters are translated into that same tree structure,
to have a uniform processing of the request.
"""
